"""
Agno Tools implemented using the Synvya SDK
"""

from .buyer import BuyerTools
from .dadjokegamer import DadJokeGamerTools
from .seller import MerchantTools

__all__ = ["BuyerTools", "MerchantTools", "DadJokeGamerTools"]



import logging
from typing import ClassVar, List, Optional, Set, Union

from agno.knowledge.knowledge import Knowledge
from agno.tools import Toolkit
from synvya_sdk import (
    ClassifiedListing,
    NostrClient,
    Product,
    Profile,
    ProfileFilter,
    Stall,
)

class BuyerTools(Toolkit):
    # Class variables
    _instances_from_create: ClassVar[Set[int]]
    merchants: Set[Profile]

    # Instance variables
    relays: List[str]
    private_key: str
    knowledge_base: Knowledge
    _nostr_client: Optional[NostrClient]
    profile: Optional[Profile]
    _instance_id: int

    # Initialization
    def __init__(
        self,
        knowledge_base: Knowledge,
        relays: Union[str, List[str]],
        private_key: str,
        _from_create: bool = False,
    ) -> None: ...
    def __del__(self) -> None: ...
    @classmethod
    async def create(
        cls,
        knowledge_base: Knowledge,
        relays: Union[str, List[str]],
        private_key: str,
        log_level: Optional[int] = logging.INFO,
    ) -> "BuyerTools": ...
    def get_profile(self) -> str: ...
    def get_relay(self) -> str: ...
    def get_relays(self) -> List[str]: ...
    async def async_set_profile(self, profile: Profile) -> str: ...

    # Retrieve NIP-15 Marketplace information from Nostr
    # and store it in the local knowledge base
    async def async_get_merchants(
        self, profile_filter_json: Optional[str | dict] = None
    ) -> str: ...
    async def async_get_merchants_in_marketplace(
        self,
        owner_public_key: str,
        name: str,
        profile_filter: Optional[ProfileFilter] = None,
    ) -> str: ...
    async def async_get_products(
        self, merchant_public_key: str, stall: Optional[Stall] = None
    ) -> str: ...
    async def async_get_stalls(self, merchant_public_key: str) -> str: ...
    async def async_get_classified_listings(
        self, profile_filter_json: Optional[str | dict] = None
    ) -> str: ...

    # Query information from local knowledge base
    def get_merchants_from_knowledge_base(
        self, search_query: str, profile_filter_json: Optional[str | dict] = None
    ) -> str: ...
    def get_products_from_knowledge_base(
        self,
        merchant_public_key: Optional[str] = None,
        categories: Optional[List[str]] = None,
    ) -> str: ...
    def get_stalls_from_knowledge_base(
        self, merchant_public_key: Optional[str] = None
    ) -> str: ...
    def get_classified_listings_from_knowledge_base(
        self,
        merchant_public_key: Optional[str] = None,
        categories: Optional[List[str]] = None,
    ) -> str: ...

    # Order products
    async def async_submit_order(self, product_name: str, quantity: int) -> str: ...
    async def async_listen_for_message(self, timeout: int = 5) -> str: ...
    async def async_submit_payment(self, payment_request: str) -> str: ...
    # Internal methods
    def _create_customer_order(
        self,
        product_id: str,
        quantity: int,
        shipping_id: str,
        address: Optional[str] = None,
    ) -> str: ...
    def _get_product_from_kb(self, product_name: str) -> Product: ...
    async def _store_profile_in_kb(self, profile: Profile) -> None: ...
    def _store_product_in_kb(self, product: Product) -> None: ...
    def _store_classified_listing_in_kb(self, listing: ClassifiedListing) -> None: ...
    def _store_stall_in_kb(self, stall: Stall) -> None: ...
    def _message_is_payment_request(self, message: str) -> bool: ...
    def _message_is_payment_verification(self, message: str) -> bool: ...
    @staticmethod
    def _normalize_hashtag(tag: str) -> str: ...



"""
Module implementing the MerchantTools Toolkit for Agno agents.
"""

import json
import time
from typing import List, Optional, Tuple, Union

from nostr_sdk import EventId
from pydantic import ConfigDict

from synvya_sdk import NostrClient, Product, Profile, Stall

try:
    from agno.tools import Toolkit
    from agno.utils.log import logger
except ImportError as exc:
    raise ImportError(
        "Package `agno` not installed. Please install using `pip install agno`"
    ) from exc


class MerchantTools(Toolkit):
    """
    MerchantTools is a toolkit that allows a merchant to publish
    products and stalls to Nostr.

    TBD:
    - At initialization, products and stalls are added to the internal database
    but not published to Nostr. User must call publish_products() and
    publish_stalls() to publish the products and stalls to Nostr.

    After the initial publication, the user can call get_products() and
    get_stalls() to get the current products and stalls in the internal database, update
    the products or stalls, and then call set_products() and set_stalls() to
    update the internal database and simultaneously publish the changes to Nostr.
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True, extra="allow", validate_assignment=True
    )

    _instances_from_create: set[int] = set()

    def __init__(
        self,
        relays: Union[str, List[str]],
        private_key: str,
        stalls: List[Stall],
        products: List[Product],
        _from_create: bool = False,
    ):
        """
        Initialize MerchantTools with a private key and a relay.
        """
        if not _from_create:
            raise RuntimeError(
                "MerchantTools must be created using the create() method"
            )

        # Track instance ID
        self._instance_id = id(self)
        MerchantTools._instances_from_create.add(self._instance_id)

        # Initialize Toolkit base class with name
        super().__init__(name="merchant")

        # Set include_tools to None to allow all tools to be registered
        self.include_tools = None

        # Convert single relay to list for consistent handling
        self.relays: List[str] = [relays] if isinstance(relays, str) else relays
        self.private_key: str = private_key

        self.stall_db: List[Tuple[Stall, Optional[str]]] = [(s, None) for s in stalls]
        self.product_db: List[Tuple[Product, Optional[str]]] = [
            (p, None) for p in products
        ]

        self.nostr_client: Optional[NostrClient] = None
        self.profile: Optional[Profile] = None

        # Register methods
        self.register(self.get_profile)
        self.register(self.get_products)
        self.register(self.get_relay)
        self.register(self.get_relays)
        self.register(self.get_stalls)
        self.register(self.async_listen_for_orders)
        self.register(self.manual_order_workflow)
        self.register(self.async_publish_product)
        self.register(self.async_publish_products)
        self.register(self.async_publish_stall)
        self.register(self.async_publish_stalls)
        self.register(self.async_send_payment_request)
        self.register(self.async_send_payment_verification)
        self.register(self.async_set_products)
        self.register(self.async_set_profile)
        self.register(self.async_set_stalls)
        self.register(self.async_remove_products)
        self.register(self.async_remove_stalls)
        self.register(self.verify_payment)

    def __del__(self) -> None:
        """
        Delete the MerchantTools instance.
        """
        if hasattr(self, "_instance_id"):
            MerchantTools._instances_from_create.discard(self._instance_id)

    @classmethod
    async def create(
        cls,
        relays: Union[str, List[str]],
        private_key: str,
        stalls: List[Stall],
        products: List[Product],
    ) -> "MerchantTools":
        """
        Create a new MerchantTools instance

        Args:
            relays: Nostr relay(s) that the client will connect to. Can be a single URL string or a list of URLs.
            private_key: Private key for the client in hex or bech32 format
            stalls: List of stalls to manage
            products: List of products to manage

        Returns:
            MerchantTools: An initialized MerchantTools instance
        """
        instance = cls(relays, private_key, stalls, products, _from_create=True)

        instance.nostr_client = await NostrClient.create(relays, private_key)
        instance.nostr_client.set_logging_level(logger.getEffectiveLevel())
        instance.profile = await instance.nostr_client.async_get_profile()
        return instance

    def get_profile(self) -> str:
        """
        Get the merchant profile in JSON format

        Returns:
            str: merchant profile in JSON format
        """
        if self.profile is None:
            raise ValueError("Profile not initialized. Please use create() method.")

        return json.dumps(self.profile.to_json())

    def get_products(self) -> str:
        """
        Get all the merchant products

        Returns:
            str: JSON string containing all products
        """
        return json.dumps([p.to_dict() for p, _ in self.product_db])

    def get_relay(self) -> str:
        """
        Get the Nostr relay the merchant is using

        Returns:
            str: Primary Nostr relay (first in the list)
        """
        return self.relays[0] if self.relays else ""

    def get_relays(self) -> List[str]:
        """
        Get all Nostr relays the merchant is using

        Returns:
            List[str]: List of Nostr relays
        """
        return self.relays

    def get_stalls(self) -> str:
        """
        Get all the merchant stalls in JSON format

        Returns:
            str: JSON string containing all stalls
        """
        return json.dumps([s.to_dict() for s, _ in self.stall_db])

    async def async_listen_for_orders(self, timeout: int = 5) -> str:
        """
        Listens for incoming orders from the Nostr relay.
        Returns one order in JSON format.

        If there are no orders, returns a JSON string with
        {"type": "none", "content": "No orders received after <timeout> seconds"}

        Args:
            timeout: timeout for the listen operation

        Returns:
            str: JSON string
            {
                "type": "order", "none",
                "kind": "kind:4", "kind:14", "none",
                "buyer": "<buyer bech32 public key>", "none",
                "content": "<order content>"
            }


        Raises:
            RuntimeError: if unable to listen for private messages
        """
        if self.nostr_client is None:
            logger.error("NostrClient not initialized. Please use create() method.")
            raise ValueError("NostrClient not initialized. Please use create() method.")

        try:
            message = await self.nostr_client.async_receive_message(timeout)
            message_dict = json.loads(message)
            message_kind = message_dict.get("type")
            if message_kind in ("kind:4", "kind:14"):
                if self._message_is_order(message_dict.get("content")):
                    return json.dumps(
                        {
                            "type": "order",
                            "kind": message_kind,
                            "buyer": message_dict.get("sender"),
                            "content": message_dict.get("content"),
                        }
                    )
            return json.dumps(
                {
                    "type": "none",
                    "kind": "none",
                    "buyer": "none",
                    "content": f"No orders received after {timeout} seconds",
                }
            )
        except RuntimeError as e:
            logger.error("Unable to listen for messages. Error %s", e)
            raise e

    def manual_order_workflow(self, buyer: str, order: str, parameters: str) -> str:
        """
        Placeholder for a manual order workflow

        Args:
            buyer: buyer bech32 public key
            order: JSON string of the order
            parameters: JSON string of the parameters

        Returns:
            str: JSON string of the payment request
        """
        return json.dumps(
            {
                "status": "success",
                "message": f"Workflow triggered for order: {order} from {buyer} with parameters: {parameters}",
            }
        )

    async def async_send_payment_request(
        self, buyer: str, order: str, kind: str, payment_type: str, payment_url: str
    ) -> str:
        """
        Processes an order

        Args:
            buyer: buyer bech32 public key
            order: JSON string of the order
            kind: message kind to use (kind:4 or kind:14)
            payment_type: Type of payment
            payment_url: URL of the payment

        Returns:
            str: JSON string of the payment request
        """
        if self.nostr_client is None:
            logger.error("NostrClient not initialized. Please use create() method.")
            raise ValueError("NostrClient not initialized. Please use create() method.")

        logger.info("process_order: Processing order: %s", order)
        try:
            order_dict = json.loads(order)
        except json.JSONDecodeError:
            return json.dumps({"status": "error", "message": "Invalid order format"})
        order_id = order_dict.get("id")

        is_valid_payment_type = payment_type in ["URL", "BTC", "LN", "LNURL"]
        if not is_valid_payment_type:
            return json.dumps({"status": "error", "message": "Invalid payment type"})

        payment_request = self._create_payment_request(
            order_id, payment_type, payment_url
        )
        response = await self.nostr_client.async_send_message(
            kind,
            buyer,
            payment_request,
        )
        return json.dumps(response)

    async def async_publish_product(self, product_name: str) -> str:
        """
        Publishes to Nostr a single product.

        Args:
            product: string with name of product

        Returns:
            str: JSON string with status of the operation

        Raises:
            ValueError: if NostrClient is not initialized
        """
        if self.nostr_client is None:
            logger.error("NostrClient not initialized. Please use create() method.")
            raise ValueError("NostrClient not initialized. Please use create() method.")

        # let's find the product
        product = next((p for p, _ in self.product_db if p.name == product_name), None)
        if product is None:
            return json.dumps(
                {"status": "error", "message": f"Product {product_name} not found"}
            )

        try:
            event_id = await self.nostr_client.async_set_product(product)
            if product not in self.product_db:
                # we need to add the product event id to the product db
                self.product_db.append((product, event_id))
            else:
                # we need to update the product event id in the product db
                self.product_db[self.product_db.index((product, None))] = (
                    product,
                    event_id,
                )
            return json.dumps(
                {
                    "status": "success",
                    "event_id": str(event_id),
                    "product_name": product.name,
                }
            )
        except RuntimeError as e:
            logger.error(
                "Failed to publish product '%s' (ID: %s): %s",
                product.name,
                product.id,
                str(e),
            )
            # Include more useful information in the error response
            return json.dumps(
                {
                    "status": "error",
                    "message": str(e),
                    "product_name": product.name,
                    "product_id": product.id,
                    "stall_id": product.stall_id,
                    "details": "Relay may be unreachable or rejecting this product format",
                }
            )

    async def async_publish_products(
        self,
        stall: Optional[Stall] = None,
        products: Optional[List[Product]] = None,
    ) -> str:
        """
        Publishes to Nostr all products in the Merchant's Product DB
        Optional parameters can be used to publish only a subset of the Product DB

        Args:
            stall: Optional publish all products for a given stall
            products: Optional subset of products to publish

        Returns:
            str: JSON array with status of all product publishing operations

        Raises:
            ValueError: if NostrClient is not initialized
        """

        if self.nostr_client is None:
            logger.error("NostrClient not initialized. Please use create() method.")
            raise ValueError("NostrClient not initialized. Please use create() method.")

        results = []

        for i, (product, _) in enumerate(self.product_db):
            if stall is not None and product.stall_id != stall.id:
                continue
            if products is not None and product not in products:
                continue
            try:
                event_id = await self.nostr_client.async_set_product(product)
                logger.debug(
                    f"Published product {product.name} with categories {', '.join(product.categories)}"
                )

                self.product_db[i] = (product, event_id)
                results.append(
                    {
                        "status": "success",
                        "event_id": str(event_id),
                        "product_name": product.name,
                    }
                )
                # Pause for 1 seconds to rate limiting by relay
                time.sleep(1)
            except RuntimeError as e:
                logger.error("Unable to publish product %s. Error %s", product, e)
                results.append(
                    {"status": "error", "message": str(e), "product_name": product.name}
                )

        return json.dumps(results)

    async def async_publish_stall(self, stall_name: str) -> str:
        """
        Publishes to Nostr a single stall.
        Adds the stall to the Stall DB if it is not already present.
        Updates the Stall DB with new event id if the stall is already present.

        Does NOT publish the products for the stall.
        Call publish_products() to publish the products for the stall.

        Args:
            stall_name: name of the stall to be published

        Returns:
            str: JSON string with status of the operation

        Raises:
            ValueError: if NostrClient is not initialized
        """
        if self.nostr_client is None:
            logger.error("NostrClient not initialized. Please use create() method.")
            raise ValueError("NostrClient not initialized. Please use create() method.")

        # let's find the stall
        stall = next((s for s, _ in self.stall_db if s.name == stall_name), None)
        if stall is None:
            return json.dumps(
                {"status": "error", "message": f"Stall {stall_name} not found"}
            )

        try:
            event_id = await self.nostr_client.async_set_stall(stall)
            if stall not in self.stall_db:
                # we need to add the stall event id to the stall db
                self.stall_db.append((stall, event_id))
            else:
                # we need to update the stall event id in the stall db
                self.stall_db[self.stall_db.index((stall, None))] = (stall, event_id)
            return json.dumps(
                {
                    "status": "success",
                    "event_id": str(event_id),
                    "stall_name": stall.name,
                }
            )
        except RuntimeError as e:
            logger.error("Unable to publish the stall: %s", e)
            return json.dumps(
                {"status": "error", "message": str(e), "stall_name": stall.name}
            )

    async def async_publish_stalls(
        self,
        stalls: Optional[List[Stall]] = None,
    ) -> str:
        """
        Publishes to Nostr all stalls managed by the merchant
        Optional parameters can be used to publish only a subset of the Stall DB

        Args:
            stalls: Optional subset of stalls to publish

        Returns:
            str: JSON array with status of all stall publishing operations

        Raises:
            ValueError: if NostrClient is not initialized
        """
        if self.nostr_client is None:
            logger.error("NostrClient not initialized. Please use create() method.")
            raise ValueError("NostrClient not initialized. Please use create() method.")

        results = []

        for i, (stall, _) in enumerate(self.stall_db):
            if stalls is not None and stall not in stalls:
                continue
            try:
                event_id = await self.nostr_client.async_set_stall(stall)
                self.stall_db[i] = (stall, event_id)
                results.append(
                    {
                        "status": "success",
                        "event_id": str(event_id),
                        "stall_name": stall.name,
                    }
                )
                # Pause for 0.5 seconds to avoid rate limiting
                time.sleep(0.5)
            except RuntimeError as e:
                logger.error("Unable to publish stall %s. Error %s", stall, e)
                results.append(
                    {"status": "error", "message": str(e), "stall_name": stall.name}
                )

        return json.dumps(results)

    async def async_set_products(self, products: List[Product]) -> str:
        """
        Sets the products used by the Toolkit.
        The products are also published to the Nostr network.
        """
        self.product_db = [(product, None) for product in products]
        return await self.async_publish_products()

    async def async_set_profile(self, profile: Profile) -> str:
        """
        Sets the profile used by the Toolkit.
        The profile is also published to the Nostr network.

        Returns:
            str: id of the event publishing the profile

        Raises:
            RuntimeError: if it can't publish the event
        """
        if self.nostr_client is None:
            logger.error("NostrClient not initialized. Please use create() method.")
            raise ValueError("NostrClient not initialized. Please use create() method.")

        try:
            return await self.nostr_client.async_set_profile(profile)
        except RuntimeError as e:
            logger.error("Unable to publish the profile: %s", e)
            raise RuntimeError(f"Unable to publish the profile: {e}") from e

    async def async_set_stalls(self, stalls: List[Stall]) -> str:
        """
        Sets the stalls used by the Toolkit.
        The stalls are also published to the Nostr network.
        """
        self.stall_db = [(stall, None) for stall in stalls]
        return await self.async_publish_stalls()

    async def async_remove_products(
        self,
        stall: Optional[Stall] = None,
        products: Optional[List[Product]] = None,
    ) -> str:
        """
        Removes from Nostr and the internal database all products published by the merchant
        Optional parameters can be used to remove only a subset of the products

        Args:
            stall: Optional remove only products for a given stall
            products: Optional subset of products to remove

        Returns:
            str: JSON array with status of all product removal operations

        Raises:
            ValueError: if NostrClient is not initialized
        """
        if self.nostr_client is None:
            logger.error("NostrClient not initialized. Please use create() method.")
            raise ValueError("NostrClient not initialized. Please use create() method.")

        results = []

        for product, event_id in self.product_db:
            if stall is not None and product.stall_id != stall.id:
                continue
            if products is not None and product not in products:
                continue

            if event_id is None:
                # product has not been published to Nostr
                # remove from database and call it a success
                self.product_db.remove((product, event_id))
                results.append(
                    {
                        "status": "success",
                        "message": f"Product '{product.name}' removed",
                        "product_name": product.name,
                        "event_id": "not previously published",
                    }
                )

                continue

            try:
                # product has been published to Nostr
                # delete the event from Nostr
                # remove from database and call it a success

                delete_event_id = await self.nostr_client.async_delete_event(
                    event_id, reason=f"Product '{product.name}' removed"
                )
                self.product_db.remove((product, event_id))
                results.append(
                    {
                        "status": "success",
                        "message": f"Product '{product.name}' removed",
                        "product_name": product.name,
                        "event_id": str(delete_event_id),
                    }
                )
                # Pause for 0.5 seconds to avoid rate limiting
                time.sleep(0.5)
            except RuntimeError as e:
                logger.error("Unable to remove product %s. Error %s", product, e)
                results.append(
                    {"status": "error", "message": str(e), "product_name": product.name}
                )

        return json.dumps(results)

    async def async_remove_stalls(
        self,
        stalls: Optional[List[Stall]] = None,
    ) -> str:
        """
        Removes from Nostr and the internal database all stalls managed by the merchant
        Optional parameters can be used to remove only a subset of the stalls

        Does ALSO remove all products within the stalls.

        Args:
            stalls: Optional subset of stalls to remove

        Returns:
            str: JSON array with status of all removal operations

        Raises:
            ValueError: if NostrClient is not initialized
        """
        if self.nostr_client is None:
            logger.error("NostrClient not initialized. Please use create() method.")
            raise ValueError("NostrClient not initialized. Please use create() method.")

        results: list[dict[str, str]] = []

        for stall, event_id in self.stall_db:
            if stalls is not None and stall not in stalls:
                continue  # we're filtering out the stalls that are not in the list

            # remove all products in this stall
            product_results = json.loads(
                await self.async_remove_products(stall=stall, products=None)
            )
            results.extend(product_results)

            # now remove the stall
            try:
                delete_event_id = await self.nostr_client.async_delete_event(
                    stall.id, reason=f"Stall '{stall.name}' removed"
                )
                self.stall_db.remove((stall, event_id))
                results.append(
                    {
                        "status": "success",
                        "message": f"Stall '{stall.name}' removed",
                        "stall_name": stall.name,
                        "event_id": str(delete_event_id),
                    }
                )

                # Pause for 0.5 seconds to avoid rate limiting
                time.sleep(0.5)
            except RuntimeError as e:
                logger.error("Unable to remove stall %s. Error %s", stall.name, e)
                results.append(
                    {"status": "error", "message": str(e), "stall_name": stall.name}
                )

        return json.dumps(results)

    def verify_payment(
        self,
        buyer: str,
        order: str,
        kind: str,
        payment_type: str,
        payment_url: str,
    ) -> str:
        """
        Verifies that payment has been received for an order
        Assumes that payment has already been received
        Sends a payment verification to the buyer
        """
        return json.dumps(
            {
                "status": "success",
                "message": "Payment verified",
            }
        )

    async def async_send_payment_verification(
        self, buyer: str, order: str, kind: str
    ) -> str:
        """
        Verifies that payment has been received for an order
        Assumes that payment has already been received
        Sends a payment verification to the buyer

        TBD: confirm payment is received

        Args:
            buyer: Bech32 public key of the buyer
            order: JSON string of the order
            kind: message kind to use (kind:4 or kind:14)

        Returns:
            str: JSON string of the payment request
        """
        if self.nostr_client is None:
            logger.error("NostrClient not initialized. Please use create() method.")
            raise ValueError("NostrClient not initialized. Please use create() method.")

        logger.info(
            "send_payment_verification: Sending payment verification for order: %s",
            order,
        )
        try:
            order_dict = json.loads(order)
        except json.JSONDecodeError:
            return json.dumps({"status": "error", "message": "Invalid order format"})
        order_id = order_dict.get("id")

        payment_verification = self._create_payment_verification(order_id)
        response = await self.nostr_client.async_send_message(
            kind,
            buyer,
            payment_verification,
        )
        return json.dumps(response)

    def _message_is_order(self, message: str) -> bool:
        """
        Check if the message contains an order.

        Args:
            message (str): The message to check.

        Returns:
            bool: True if the message contains an order, False otherwise
        """
        try:
            # Check if message is already a dictionary
            if isinstance(message, dict):
                content = message
            else:
                content = json.loads(message)

            logger.debug("_message_is_order: content: %s", content)

            if content.get("type") != 0:
                return False

            items = content.get("items", [])
            if isinstance(items, list) and any(
                isinstance(item, dict) and "product_id" in item and "quantity" in item
                for item in items
            ):
                return True
            return False
        except json.JSONDecodeError:
            return False

    def _create_payment_request(
        self,
        order_id: str,
        payment_type: str,
        payment_url: str,
    ) -> str:
        """
        Create a payment request in JSON format.

        Args:
            order_id (str): ID of the order.
            payment_type (PaymentType): Type of payment.
            payment_url (str): URL of the payment.

        Returns:
            str: JSON string of the payment request.
        """

        payment_request = {
            "id": order_id,
            "type": 1,
            "message": "Thank you for your business!",
            "payment_options": [{"type": payment_type.lower(), "link": payment_url}],
        }

        return json.dumps(
            payment_request, indent=2
        )  # Convert to JSON string with pretty printing

    def _create_payment_verification(
        self,
        order_id: str,
    ) -> str:
        """
        Create a payment verification in JSON format.

        Args:
            order_id (str): ID of the order.

        Returns:
            str: JSON string of the payment request.
        """

        payment_verification = {
            "id": order_id,
            "type": 2,
            "message": "Thank you for your business!",
            "paid": True,
            "shipped": True,
        }

        return json.dumps(
            payment_verification, indent=2
        )  # Convert to JSON string with pretty printing



"""
Module implementing the BuyerTools Toolkit for Agno agents.
"""

import hashlib
import json
import logging
import re
import secrets
from sys import stdout
from typing import Any, List, Optional, Union, cast

from pydantic import ConfigDict

from synvya_sdk import (
    ClassifiedListing,
    KeyEncoding,
    Namespace,
    NostrClient,
    Product,
    Profile,
    ProfileFilter,
    ProfileType,
    Stall,
)

try:
    from agno.knowledge.document import Document
    from agno.knowledge.knowledge import Knowledge
    from agno.tools import Toolkit
    from agno.utils.log import logger
    from agno.vectordb import VectorDb
except ImportError as exc:
    raise ImportError(
        "`agno` not installed. Please install using `pip install agno`"
    ) from exc

# Create a direct console logger that doesn't rely on agno's logger
buyer_logger = logging.getLogger("synvya_buyer")
buyer_logger.setLevel(logging.INFO)
if not buyer_logger.handlers:
    handler = logging.StreamHandler(stdout)
    handler.setFormatter(logging.Formatter("%(name)s: %(message)s"))
    buyer_logger.addHandler(handler)


def _map_location_to_geohash(location: str) -> str:
    """
    Map a location to a geohash.

    TBD: Implement this function. Returning a fixed geohash for now.

    Args:
        location: location to map to a geohash. Can be a zip code, city,
        state, country, or latitude and longitude.

    Returns:
        str: geohash of the location or empty string if location is not found
    """
    if "snoqualmie" in location.lower():
        return "C23Q7U36W"

    return ""


def _get_vector_db(knowledge: Knowledge) -> Optional[VectorDb]:
    """
    Safely extract the configured vector database from a Knowledge instance.

    Args:
        knowledge: Knowledge container that may hold a vector DB.

    Returns:
        Optional[VectorDb]: The vector database if configured, otherwise None.
    """
    vector_db = getattr(knowledge, "vector_db", None)
    return cast(Optional[VectorDb], vector_db)


class BuyerTools(Toolkit):
    """
    BuyerTools is a toolkit that allows an agent to find sellers and
    transact with them over Nostr.

    `Download` tools download data from the Nostr relay and store it in the
    knowledge base.

    `Get` tools retrieve data from the knowledge base.

    TBD: populate the sellers locations with info from stalls.
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True, extra="allow", validate_assignment=True
    )

    _instances_from_create: set[int] = set()
    merchants: set[Profile]

    def __init__(
        self,
        knowledge_base: Knowledge,
        relays: Union[str, List[str]],
        private_key: str,
        _from_create: bool = False,
    ) -> None:
        """
        Initialize the BuyerTools object.

        Args:
            knowledge_base: Knowledge base to store and retrieve information
            relays: Nostr relay(s) that the client will connect to. Can be a single URL string or a list of URLs.
            private_key: Private key for the client in hex or bech32 format
            _from_create: Internal flag to ensure proper initialization flow
        """
        if not _from_create:
            raise RuntimeError("BuyerTools must be created using the create() method")

        # Track instance ID
        self._instance_id = id(self)
        BuyerTools._instances_from_create.add(self._instance_id)

        super().__init__(name="buyer")

        # Set include_tools to None to allow all tools to be registered
        self.include_tools = None

        # Convert single relay to list for consistent handling
        self.relays: List[str] = [relays] if isinstance(relays, str) else relays
        self.private_key: str = private_key
        self.knowledge_base: Knowledge = knowledge_base
        self.profile: Optional[Profile] = None
        self._nostr_client: Optional[NostrClient] = None
        BuyerTools.merchants = set()

        # Register methods
        self.register(self.async_get_merchants)
        self.register(self.get_merchants_from_knowledge_base)
        self.register(self.async_get_merchants_in_marketplace)
        self.register(self.async_get_products)
        self.register(self.async_get_classified_listings)
        self.register(self.get_products_from_knowledge_base)
        self.register(self.get_classified_listings_from_knowledge_base)
        # self.register(self.get_products_from_knowledge_base_by_category)
        self.register(self.get_profile)
        self.register(self.get_relay)
        self.register(self.async_get_stalls)
        self.register(self.get_stalls_from_knowledge_base)
        self.register(self.async_listen_for_message)
        self.register(self.async_submit_order)
        self.register(self.async_submit_payment)

    def __del__(self) -> None:
        """
        Delete the BuyerTools instance.
        """
        if hasattr(self, "_instance_id"):
            BuyerTools._instances_from_create.discard(self._instance_id)

    @classmethod
    async def create(
        cls,
        knowledge_base: Knowledge,
        relays: Union[str, List[str]],
        private_key: str,
        log_level: Optional[int] = logging.INFO,
    ) -> "BuyerTools":
        """
        Create a new BuyerTools instance

        Args:
            knowledge_base: Knowledge base to store and retrieve information
            relays: Nostr relay(s) that the client will connect to. Can be a single URL string or a list of URLs.
            private_key: Private key for the client in hex or bech32 format
            log_level: Optional logging level

        Returns:
            BuyerTools: An initialized BuyerTools instance
        """
        instance = cls(knowledge_base, relays, private_key, _from_create=True)

        # Set logging level
        if log_level is not None:
            buyer_logger.setLevel(log_level)
            NostrClient.set_logging_level(
                log_level
            )  # Set the class-level logger BEFORE creating any instances
        else:
            buyer_logger.setLevel(logger.getEffectiveLevel())
            NostrClient.set_logging_level(logger.getEffectiveLevel())

        # Then initialize NostrClient with proper logging already set up
        instance._nostr_client = await NostrClient.create(relays, private_key)

        instance.profile = await instance._nostr_client.async_get_profile()

        return instance

    async def async_get_merchants(
        self, profile_filter_json: Optional[str | dict] = None
    ) -> str:
        """
        Download from the Nostr relay all merchants and store their Nostr
        profile in the knowledge base.

        Args:
            profile_filter_json: JSON string or dictionary representing a filter to apply to merchants.
                                Format: {"namespace": "MERCHANT", "profile_type": "restaurant", "hashtags": ["pizza"]}

        Returns:
            str: JSON string with status and count of merchants refreshed
        """
        buyer_logger.debug(
            "GET_MERCHANTS: profile_filter_json: %s", profile_filter_json
        )

        # If there is no filter, get all merchants
        if profile_filter_json is None:
            try:
                self.merchants = await self._nostr_client.async_get_merchants()
            except RuntimeError as e:
                logger.error("Error downloading merchants from the Nostr relay: %s", e)
                return json.dumps({"status": "error", "message": str(e)})

            # Store merchants in knowledge base
            for merchant in self.merchants:
                await self._store_profile_in_kb(merchant)

            response = json.dumps({"status": "success", "count": len(self.merchants)})
            buyer_logger.debug("GET_MERCHANTS: response: %s", response)
            return response

        # If there is a filter, get the merchants that match the filter
        filter_data = None
        if isinstance(profile_filter_json, str):
            # Parse the JSON string into a dict
            filter_data = json.loads(profile_filter_json)
        elif isinstance(profile_filter_json, dict):
            # Use the dictionary directly
            filter_data = profile_filter_json

        # Extract the values
        namespace_str = filter_data.get("namespace")
        profile_type_str = filter_data.get("profile_type")
        hashtags = filter_data.get("hashtags", [])

        # Convert namespace string to Namespace enum
        namespace = None
        if namespace_str:
            try:
                namespace = getattr(Namespace, namespace_str)
            except AttributeError as e:
                if namespace_str in [n.value for n in Namespace]:
                    # If the string is the value rather than the name
                    namespace = [n for n in Namespace if n.value == namespace_str][0]
                else:
                    raise ValueError(f"Invalid namespace: {namespace_str}") from e

        # Convert profile_type string to ProfileType enum
        profile_type = None
        if profile_type_str:
            # Try to match by direct enum name first
            try:
                profile_type = getattr(
                    ProfileType, f"MERCHANT_{profile_type_str.upper()}"
                )
            except AttributeError as e:
                # Try to match by enum value
                matching_types = [
                    pt for pt in ProfileType if pt.value == profile_type_str
                ]
                if matching_types:
                    profile_type = matching_types[0]
                else:
                    raise ValueError(
                        f"Invalid profile_type: {profile_type_str}. "
                        f"Valid values are: {[pt.value for pt in ProfileType]}"
                    ) from e

        # Create the ProfileFilter
        profile_filter = ProfileFilter(
            namespace=namespace,
            profile_type=profile_type,
            hashtags=hashtags,
        )
        buyer_logger.debug("Created ProfileFilter: %s", profile_filter)

        # Get the merchants that match the filter
        try:
            self.merchants = await self._nostr_client.async_get_merchants(
                profile_filter
            )
        except RuntimeError as e:
            buyer_logger.error(
                "Error downloading merchants from the Nostr relay: %s", e
            )
            return json.dumps({"status": "error", "message": str(e)})

        # Store merchants in knowledge base
        for merchant in self.merchants:
            await self._store_profile_in_kb(merchant)

        response = json.dumps({"status": "success", "count": len(self.merchants)})
        buyer_logger.debug("GET_MERCHANTS: response: %s", response)

        return response

    def get_merchants_from_knowledge_base(
        self, search_query: str, profile_filter_json: Optional[str | dict] = None
    ) -> str:
        """
        Get the list of merchants stored in the knowledge base.

        Args:
            profile_filter_json: JSON string or dictionary representing a filter to apply to merchants.
                             Format: {"namespace": "com.synvya.merchant", "profile_type": "restaurant", "hashtags": ["pizza"]}

        Returns:
            str: JSON string of merchants
        """
        buyer_logger.debug(
            "GET_MERCHANTS_FROM_KNOWLEDGE_BASE: query: %s, profile_filter_json: %s",
            search_query,
            str(profile_filter_json),
        )

        # Initialize filter map used by the vector store
        search_filters: dict[str, Any] = {}

        # Process profile filter if provided
        if profile_filter_json:
            try:
                # Parse filter data
                filter_data = (
                    json.loads(profile_filter_json)
                    if isinstance(profile_filter_json, str)
                    else profile_filter_json
                )

                # Add namespace filter (exact match)
                namespace = filter_data.get("namespace")
                if namespace:
                    search_filters["namespace"] = namespace

                # Add profile_type filter (exact match)
                profile_type = filter_data.get("profile_type")
                if profile_type:
                    search_filters["profile_type"] = profile_type

                # Add hashtag filters (boolean match per hashtag)
                hashtags = filter_data.get("hashtags", [])
                for tag in hashtags:
                    normalized_tag = self._normalize_hashtag(tag)
                    search_filters[f"hashtag_{normalized_tag}"] = True

                buyer_logger.debug("Applied search filters: %s", search_filters)

            except json.JSONDecodeError as e:
                buyer_logger.error("Invalid JSON format for profile_filter: %s", e)
                return json.dumps(
                    {"status": "error", "message": f"Invalid JSON format: {str(e)}"}
                )
            except Exception as e:
                buyer_logger.error("Error processing profile filter: %s", e)
                return json.dumps(
                    {"status": "error", "message": f"Error processing filter: {str(e)}"}
                )

        buyer_logger.debug("Search filters: %s", str(search_filters))

        # Execute search
        documents = self.knowledge_base.search(
            query=search_query, max_results=100, filters=search_filters or None
        )

        buyer_logger.debug("Found %d merchants in the knowledge base", len(documents))

        # Return JSON content of found merchants
        merchants_json = [doc.content for doc in documents]
        buyer_logger.debug("Merchants JSON: %s", str(merchants_json))
        return json.dumps(merchants_json)

    async def async_get_merchants_in_marketplace(
        self,
        owner_public_key: str,
        name: str,
        profile_filter: Optional[ProfileFilter] = None,
    ) -> str:
        """
        Download from the Nostr relay all merchants included in a Nostr
        marketplace and store their Nostr profile in the knowledge base.

        Args:
            owner_public_key: bech32 encoded public key of the owner of the marketplace
            name: name of the marketplace to download merchants from

        Returns:
            str: JSON string with status and count of merchants downloaded

        TBD: Implement profile filter.
        """
        buyer_logger.debug("Downloading merchants from the Nostr marketplace %s", name)
        try:
            # Retrieve merchants from the Nostr marketplace
            self.merchants = (
                await self._nostr_client.async_get_merchants_in_marketplace(
                    owner_public_key, name, profile_filter
                )
            )
            # Store merchants in the knowledge base
            for merchant in self.merchants:
                await self._store_profile_in_kb(merchant)

            # Return the number of merchants downloaded
            response = json.dumps({"status": "success", "count": len(self.merchants)})
        except RuntimeError as e:
            buyer_logger.error(
                "Error downloading merchants from the Nostr marketplace %s: %s",
                name,
                e,
            )
            response = json.dumps({"status": "error", "message": str(e)})

        return response

    async def async_get_products(
        self, merchant_public_key: str, stall: Optional[Stall] = None
    ) -> str:
        """
        Download all products published by a merchant on Nostr and store them
        in the knowledge base.

        Args:
            merchant_public_key: public key of the merchant
            stall: optional stall to filter products by

        Returns:
            str: JSON string with all products published by the merchant
        """
        buyer_logger.debug("Downloading products from merchant %s", merchant_public_key)
        try:
            # retrieve products from the Nostr relay
            products = await self._nostr_client.async_get_products(
                merchant_public_key, stall
            )

            # store products in the knowledge base
            for product in products:
                self._store_product_in_kb(product)

            response = json.dumps([product.to_dict() for product in products])

        except RuntimeError as e:
            buyer_logger.error(
                "Error downloading products from merchant %s: %s",
                merchant_public_key,
                e,
            )
            response = json.dumps({"status": "error", "message": str(e)})

        return response

    async def async_get_classified_listings(
        self, profile_filter_json: Optional[str | dict] = None
    ) -> str:
        """
        Download classified listings for merchants that match the optional
        profile filter and store them in the knowledge base.

        Args:
            profile_filter_json: JSON string or dict representing the merchant profile filter.

        Returns:
            str: JSON string containing the downloaded classified listings.
        """
        buyer_logger.debug(
            "Downloading classified listings with profile filter: %s",
            profile_filter_json,
        )

        if self._nostr_client is None:
            buyer_logger.error("Nostr client not initialized")
            return json.dumps(
                {
                    "status": "error",
                    "message": "Nostr client not initialized",
                }
            )

        merchants_response_str = await self.async_get_merchants(profile_filter_json)
        try:
            merchants_response = json.loads(merchants_response_str)
        except json.JSONDecodeError as exc:
            buyer_logger.error(
                "Invalid response when retrieving merchants for classifieds: %s",
                exc,
            )
            return json.dumps(
                {
                    "status": "error",
                    "message": "Unable to decode merchants response",
                }
            )

        if (
            isinstance(merchants_response, dict)
            and merchants_response.get("status") == "error"
        ):
            buyer_logger.error(
                "Failed to download merchants before fetching classifieds: %s",
                merchants_response.get("message"),
            )
            return json.dumps(merchants_response)

        listings_payload: List[dict[str, Any]] = []
        error_merchants: List[str] = []

        for merchant in getattr(self, "merchants", set()):
            merchant_key = merchant.get_public_key()
            try:
                listings = await self._nostr_client.async_get_classified_listings(
                    merchant_key
                )
            except RuntimeError as err:
                buyer_logger.error(
                    "Error downloading classified listings from merchant %s: %s",
                    merchant_key,
                    err,
                )
                error_merchants.append(merchant_key)
                continue

            for listing in listings:
                if not listing.get_seller():
                    listing.set_seller(merchant_key)

                self._store_classified_listing_in_kb(listing)
                listings_payload.append(listing.to_dict())

        if error_merchants:
            buyer_logger.debug(
                "Failed to download classified listings for merchants: %s",
                ", ".join(error_merchants),
            )

        buyer_logger.debug(
            "Retrieved %d classified listings for profile filter %s",
            len(listings_payload),
            profile_filter_json,
        )

        return json.dumps(listings_payload)

    def get_products_from_knowledge_base(
        self,
        merchant_public_key: Optional[str] = None,
        categories: Optional[List[str]] = None,
    ) -> str:
        """
        Get a list of products stored in the knowledge base. Optionally filter by
        merchant or categories.

        Args:
            merchant_public_key: optional filter bymerchant
            categories: optional filter by the list of categories

        Returns:
            str: JSON string of products
        """
        buyer_logger.debug("Getting products from knowledge base")

        if merchant_public_key is not None:
            search_query = merchant_public_key
        else:
            search_query = ""

        if categories is not None:
            search_filters = [
                {"type": "product"},
                {"categories": categories},
            ]
        else:
            search_filters = [
                {"type": "product"},
            ]

        documents = self.knowledge_base.search(
            query=search_query, max_results=100, filters=search_filters
        )
        for doc in documents:
            buyer_logger.debug("Document: %s", doc.to_dict())

        products_json = [doc.content for doc in documents]
        buyer_logger.debug(
            "Found %d products in the knowledge base", len(products_json)
        )
        return json.dumps(products_json)

    def get_classified_listings_from_knowledge_base(
        self,
        merchant_public_key: Optional[str] = None,
        categories: Optional[List[str]] = None,
    ) -> str:
        """
        Retrieve classified listings stored in the knowledge base, optionally
        filtered by merchant or categories.

        Args:
            merchant_public_key: Optional merchant public key to filter by.
            categories: Optional list of categories to filter by.

        Returns:
            str: JSON string of classified listings.
        """
        buyer_logger.debug("Getting classified listings from knowledge base")

        search_query = merchant_public_key if merchant_public_key is not None else ""

        if categories is not None:
            search_filters = [
                {"type": "classified_listing"},
                {"categories": categories},
            ]
        else:
            search_filters = [
                {"type": "classified_listing"},
            ]

        documents = self.knowledge_base.search(
            query=search_query, max_results=100, filters=search_filters
        )
        for doc in documents:
            buyer_logger.debug("Classified document: %s", doc.to_dict())

        listings_json = [doc.content for doc in documents]
        buyer_logger.debug(
            "Found %d classified listings in the knowledge base", len(listings_json)
        )
        return json.dumps(listings_json)

    def get_profile(self) -> str:
        """
        Get the Nostr profile of the buyer agent.

        Returns:
            str: buyer profile json string
        """
        buyer_logger.debug("Getting own profile")
        return self.profile.to_json()

    def get_relay(self) -> str:
        """Get the Nostr relay that the buyer agent is using.

        Returns:
            str: Primary Nostr relay (first in the list)
        """
        return self.relays[0] if self.relays else ""

    def get_relays(self) -> List[str]:
        """Get all Nostr relays that the buyer agent is using.

        Returns:
            List[str]: List of Nostr relays
        """
        return self.relays

    async def async_get_stalls(self, merchant_public_key: str) -> str:
        """
        Download all stalls published by a merchant on Nostr and store them
        in the knowledge base.

        Args:
            merchant_public_key: public key of the merchant

        Returns:
            str: JSON string with all stalls published by the merchant
        """
        buyer_logger.debug("Downloading stalls from merchant %s", merchant_public_key)
        try:
            # retrieve stalls from the Nostr relay
            stalls = await self._nostr_client.async_get_stalls(merchant_public_key)

            # store stalls in the knowledge base
            for stall in stalls:
                self._store_stall_in_kb(stall)

            # convert stalls to JSON string
            response = json.dumps([stall.to_dict() for stall in stalls])
        except RuntimeError as e:
            buyer_logger.error(
                "Error downloading stalls from merchant %s: %s",
                merchant_public_key,
                e,
            )
            response = json.dumps({"status": "error", "message": str(e)})

        return response

    def get_stalls_from_knowledge_base(
        self, merchant_public_key: Optional[str] = None
    ) -> str:
        """
        Get the list of stalls stored in the knowledge base.
        Optionally filter by merchant.

        Args:
            merchant_public_key: optional filter by merchant

        Returns:
            str: JSON string of stalls
        """
        buyer_logger.debug("Getting stalls from knowledge base")

        if merchant_public_key is not None:
            search_query = merchant_public_key
        else:
            search_query = ""

        documents = self.knowledge_base.search(
            query=search_query, max_results=100, filters=[{"type": "stall"}]
        )
        for doc in documents:
            buyer_logger.debug("Document: %s", doc.to_dict())

        stalls_json = [doc.content for doc in documents]
        buyer_logger.debug("Found %d stalls in the knowledge base", len(stalls_json))
        return json.dumps(stalls_json)

    async def async_listen_for_message(self, timeout: int = 5) -> str:
        """
        Listens for incoming messages from the Nostr relay.
        Returns one message in JSON format.

        Args:
            timeout: timeout for the listen operation

        Returns:
            str: JSON string
            {
                "type": "payment request", "payment verification", "unknown",
                "kind": "kind:4", "kind:14", "none",
                "seller": "<seller bech32 public key>", "none",
                "content": "<order content>"
            }


        Raises:
            RuntimeError: if unable to listen for private messages
        """
        try:
            message = await self._nostr_client.async_receive_message(timeout)
            message_dict = json.loads(message)
            message_kind = message_dict.get("type")
            if message_kind in {"kind:4", "kind:14"}:
                if self._message_is_payment_request(message_dict.get("content")):
                    return json.dumps(
                        {
                            "type": "payment request",
                            "seller": message_dict.get("sender"),
                            "content": message_dict.get("content"),
                        }
                    )
                if self._message_is_payment_verification(message_dict.get("content")):
                    return json.dumps(
                        {
                            "type": "payment verification",
                            "seller": message_dict.get("sender"),
                            "content": message_dict.get("content"),
                        }
                    )
            return json.dumps(
                {
                    "type": "unknown",
                    "kind": message_kind,
                    "buyer": "none",
                    "content": f"No orders received after {timeout} seconds",
                }
            )
        except RuntimeError as e:
            buyer_logger.error("Unable to listen for messages. Error %s", e)
            raise e

    async def async_set_profile(self, profile: Profile) -> str:
        """
        Set the Nostr profile of the buyer agent.

        Args:
            profile: Nostr profile to set

        Returns:
            str: Nostr profile json string
        """
        self.profile = profile
        try:
            await self._nostr_client.async_set_profile(profile)
        except (RuntimeError, ValueError) as e:
            buyer_logger.error("Error setting profile: %s", e)
            return json.dumps({"status": "error", "message": str(e)})

        return json.dumps({"status": "success"})

    async def async_submit_order(self, product_name: str, quantity: int) -> str:
        """
        Purchase a product.

        TBD: Complete flow. Today it just sends first message
        and returns a fixed response.

        Args:
            product_name: name of the product to purchase
            quantity: quantity of the product to purchase

        Returns:
            str: JSON string with status and message
        """

        try:
            product = self._get_product_from_kb(product_name)
        except RuntimeError as e:
            buyer_logger.error("Error getting product from knowledge base: %s", e)
            return json.dumps({"status": "error", "message": str(e)})

        if not product.get_seller():
            buyer_logger.error("Product %s has no seller", product_name)
            return json.dumps({"status": "error", "message": "Product has no seller"})

        try:
            # Confirm seller has valid NIP-05
            merchant = await self._nostr_client.async_get_profile(product.get_seller())
            if not merchant.is_nip05_validated():
                buyer_logger.error(
                    "Merchant %s does not have a verified NIP-05", product.get_seller()
                )
                return json.dumps(
                    {
                        "status": "error",
                        "message": "Merchant does not have a verified NIP-05",
                    }
                )
        except (ValueError, RuntimeError) as e:
            buyer_logger.error("Error retrieving seller profile: %s", e)
            return json.dumps(
                {
                    "status": "error",
                    "message": f"Unable to retrieve seller profile for {product.get_seller()}: {str(e)}",
                }
            )

        # Choosing the first shipping zone for now
        # Address is hardcoded for now. Add it to the buyer profile later.
        order_msg = self._create_customer_order(
            product.id,
            quantity,
            product.shipping[0].get_id(),
            "123 Main St, Anytown, USA",
        )

        await self._nostr_client.async_send_message(
            "kind:14",
            product.get_seller(),
            order_msg,
        )

        return json.dumps(
            {
                "status": "success",
                "message": f"Order placed for {quantity} units of {product_name}",
                "seller": product.get_seller(),
            }
        )

    async def async_submit_payment(self, payment_request: str) -> str:
        """
        Submit a payment to the seller.
        TBD: Complete flow. Today it just returns a fixed response.

        Args:
            payment_request: payment request to submit

        Returns:
            str: JSON string with status and message
        """
        buyer_logger.debug("Submitting payment: %s", payment_request)

        return json.dumps(
            {
                "status": "success",
                "message": "Payment submitted",
            }
        )

    def _create_customer_order(
        self,
        product_id: str,
        quantity: int,
        shipping_id: str,
        address: Optional[str] = None,
    ) -> str:
        # Use secrets module for cryptographically secure random number generation
        random_order_id = secrets.randbelow(
            1000000000
        )  # Generate a number between 0 and 999999999
        customer_order_id_str = f"{random_order_id:09d}"

        customer_order = {
            "id": customer_order_id_str,
            "type": 0,
            "name": self.profile.name,
            "address": address,
            "message": "Please accept this order.",
            "contact": {
                "nostr": self.profile.public_key,
                "phone": "",
                "email": "",
            },
            "items": [{"product_id": product_id, "quantity": quantity}],
            "shipping_id": shipping_id,
        }

        return json.dumps(
            customer_order, indent=2
        )  # Convert to JSON string with pretty printing

    def _get_product_from_kb(self, product_name: str) -> Product:
        """
        Get a product from the knowledge base.
        """
        buyer_logger.debug("Getting product from knowledge base: %s", product_name)
        documents = self.knowledge_base.search(
            query=product_name, max_results=1, filters=[{"type": "product"}]
        )
        if len(documents) == 0:
            raise RuntimeError(f"Product {product_name} not found in knowledge base")
        return Product.from_json(documents[0].content)

    def _message_is_payment_request(self, message: str) -> bool:
        """
        Check if a message is a payment request.
        Args:
            message: message to check

        Returns:
            bool: True if the message is a payment request, False otherwise

        Raises:
            json.JSONDecodeError: if the message is not a valid JSON string
        """
        try:
            # Check if message is already a dictionary
            if isinstance(message, dict):
                content = message
            else:
                content = json.loads(message)

            buyer_logger.debug("_message_is_payment_request: content: %s", content)

            if content.get("type") != 1:
                return False

            payment_options = content.get("payment_options", [])
            if isinstance(payment_options, list) and any(
                isinstance(payment_option, dict)
                and "type" in payment_option
                and "link" in payment_option
                for payment_option in payment_options
            ):
                return True
            return False
        except json.JSONDecodeError:
            return False

    def _message_is_payment_verification(self, message: str) -> bool:
        """
        Check if a message is a payment verification.
        Args:
            message: message to check

        Returns:
            bool: True if the message is a payment verification, False otherwise

        Raises:
            json.JSONDecodeError: if the message is not a valid JSON string
        """
        try:
            # Check if message is already a dictionary
            if isinstance(message, dict):
                content = message
            else:
                content = json.loads(message)

            buyer_logger.debug("_message_is_payment_verification: content: %s", content)

            if content.get("type") != 2:
                return False

            paid = content.get("paid")
            shipped = content.get("shipped")

            if isinstance(paid, bool) and isinstance(shipped, bool):
                return True
            return False

        except json.JSONDecodeError:
            return False

    async def _store_profile_in_kb(self, profile: Profile) -> None:
        """
        Store a Nostr profile directly in the vector database.

        Args:
            profile: Nostr profile to store
        """
        buyer_logger.debug("_store_profile_in_kb: profile: %s", profile.get_name())

        vector_db = _get_vector_db(self.knowledge_base)
        if vector_db is None:
            buyer_logger.warning("Vector DB not configured; skipping profile storage")
            return

        namespace = profile.get_namespace()
        profile_type = (
            profile.get_profile_type().value if profile.get_profile_type() else None
        )
        filters: dict[str, Any] = {}
        if namespace:
            filters["namespace"] = namespace
        if profile_type:
            filters["profile_type"] = profile_type

        for tag in profile.get_hashtags():
            filters[f"hashtag_{self._normalize_hashtag(tag)}"] = True

        profile_json = profile.to_json()
        if not profile_json:
            buyer_logger.warning(
                "Profile serialization returned empty payload for %s; skipping storage",
                profile.get_name(),
            )
            return

        document_id = profile.get_public_key(KeyEncoding.HEX)
        document = Document(
            id=document_id,
            name=profile.get_name(),
            content=profile_json,
            meta_data=filters.copy(),
        )
        document.content_id = document_id

        hash_source = document_id or profile_json
        content_hash = hashlib.sha256(
            hash_source.encode("utf-8", errors="ignore")
        ).hexdigest()

        try:
            await vector_db.async_upsert(content_hash, [document], filters)
        except NotImplementedError:
            vector_db.upsert(content_hash, [document], filters)
        except Exception as err:
            buyer_logger.error(
                "Failed to upsert profile %s into vector DB: %s",
                profile.get_name(),
                err,
            )
            return

        self.knowledge_base.add_filters(filters)
        # self.knowledge_base.load_document(
        #     document=doc,
        #     filters=filters,
        # )

    def _store_product_in_kb(self, product: Product) -> None:
        """
        Store a Nostr product in the knowledge base.

        Args:
            product: Nostr product to store
        """
        buyer_logger.debug("Storing product in knowledge base: %s", product.name)

        vector_db = _get_vector_db(self.knowledge_base)
        if vector_db is None:
            buyer_logger.warning("Vector DB not configured; skipping product storage")
            return

        product_json = product.to_json()
        if not product_json:
            buyer_logger.warning(
                "Product serialization returned empty payload for %s", product.name
            )
            return

        metadata: dict[str, Any] = {
            "type": "product",
            "seller": product.seller,
            "stall_id": product.stall_id,
            "categories": product.categories,
        }
        for category in product.categories:
            metadata[f"category_{self._normalize_hashtag(category)}"] = True

        document_id = f"product:{product.id}"
        document = Document(
            id=document_id,
            name=product.name,
            content=product_json,
            meta_data=metadata.copy(),
        )
        document.content_id = document_id
        content_hash = hashlib.sha256(
            document_id.encode("utf-8", errors="ignore")
        ).hexdigest()

        try:
            vector_db.upsert(content_hash, [document], metadata)
        except Exception as err:
            buyer_logger.error(
                "Failed to upsert product %s into vector DB: %s", product.id, err
            )
            return

        self.knowledge_base.add_filters(metadata)

    def _store_classified_listing_in_kb(self, listing: ClassifiedListing) -> None:
        """
        Store a classified listing in the knowledge base.

        Args:
            listing: Classified listing to store.
        """
        buyer_logger.debug(
            "Storing classified listing in knowledge base: %s", listing.title
        )

        vector_db = _get_vector_db(self.knowledge_base)
        if vector_db is None:
            buyer_logger.warning(
                "Vector DB not configured; skipping classified listing storage"
            )
            return

        if listing.location and not listing.geohash:
            computed_geohash = _map_location_to_geohash(listing.location)
            if computed_geohash:
                listing.geohash = computed_geohash

        listing_json = listing.to_json()
        if not listing_json:
            buyer_logger.warning(
                "Classified listing serialization returned empty payload for %s",
                listing.title,
            )
            return

        seller = listing.get_seller()
        metadata: dict[str, Any] = {
            "type": "classified_listing",
            "seller": seller,
            "listing_type": listing.listing_type,
            "listing_format": listing.listing_format,
            "visibility": listing.visibility,
            "categories": listing.categories,
        }
        if listing.location:
            metadata["location"] = listing.location
        if listing.geohash:
            metadata["geohash"] = listing.geohash
        if listing.collections:
            metadata["collections"] = listing.collections
        if listing.price_currency:
            metadata["price_currency"] = listing.price_currency
        for category in listing.categories:
            metadata[f"category_{self._normalize_hashtag(category)}"] = True

        document_id = f"classified_listing:{seller}:{listing.id}"
        document = Document(
            id=document_id,
            name=listing.title,
            content=listing_json,
            meta_data=metadata.copy(),
        )
        document.content_id = document_id
        content_hash = hashlib.sha256(
            document_id.encode("utf-8", errors="ignore")
        ).hexdigest()

        try:
            vector_db.upsert(content_hash, [document], metadata)
        except Exception as err:
            buyer_logger.error(
                "Failed to upsert classified listing %s into vector DB: %s",
                listing.id,
                err,
            )
            return

        self.knowledge_base.add_filters(metadata)

    def _store_stall_in_kb(self, stall: Stall) -> None:
        """
        Store a Nostr stall in the knowledge base.

        Args:
            stall: Nostr stall to store
        """
        buyer_logger.debug("Storing stall in knowledge base: %s", stall.name)

        vector_db = _get_vector_db(self.knowledge_base)
        if vector_db is None:
            buyer_logger.warning("Vector DB not configured; skipping stall storage")
            return

        stall_json = stall.to_json()
        if not stall_json:
            buyer_logger.warning(
                "Stall serialization returned empty payload for %s", stall.name
            )
            return

        shipping_methods = [method.to_dict() for method in stall.shipping]
        metadata: dict[str, Any] = {
            "type": "stall",
            "currency": stall.currency,
            "geohash": stall.get_geohash(),
            "shipping_methods": shipping_methods,
        }

        document_id = f"stall:{stall.id}"
        document = Document(
            id=document_id,
            name=stall.name,
            content=stall_json,
            meta_data=metadata.copy(),
        )
        document.content_id = document_id
        content_hash = hashlib.sha256(
            document_id.encode("utf-8", errors="ignore")
        ).hexdigest()

        try:
            vector_db.upsert(content_hash, [document], metadata)
        except Exception as err:
            buyer_logger.error(
                "Failed to upsert stall %s into vector DB: %s", stall.id, err
            )
            return

        self.knowledge_base.add_filters(metadata)

    @staticmethod
    def _normalize_hashtag(tag: str) -> str:
        """
        Normalize hashtags by removing spaces, underscores, and hyphens,
        and converting to lowercase.
        Ensures consistent matching across variations.
        """
        tag = tag.lower()
        tag = re.sub(r"[\s\-_]+", "", tag)  # Remove spaces, hyphens, underscores
        return tag



import json
import logging
import re
import warnings
from datetime import datetime, timezone
from enum import Enum
from functools import wraps
from typing import ClassVar, Dict, List, Literal, Optional, Set

import httpx
from nostr_sdk import (
    Alphabet,
    Event,
    JsonValue,
    Keys,
    Kind,
    Metadata,
    ProductData,
    PublicKey,
    ShippingCost,
    ShippingMethod,
    SingleLetterTag,
    StallData,
    Tag,
    TagKind,
)
from pydantic import BaseModel, ConfigDict, Field


def deprecated(reason: str, version: str = "2.0.0", alternative: str = None):
    """
    Decorator to mark functions as deprecated.

    Args:
        reason: Why the function is deprecated
        version: Version when it will be removed
        alternative: Suggested alternative function/method
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            message = f"{func.__name__} is deprecated. {reason}"
            if alternative:
                message += f" Use {alternative} instead."
            message += f" Will be removed in version {version}."

            warnings.warn(message, DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        # Update docstring
        if func.__doc__:
            func.__doc__ += f"\n\n.. deprecated:: {version}\n    {reason}"
            if alternative:
                func.__doc__ += f" Use :func:`{alternative}` instead."

        return wrapper

    return decorator


class KeyEncoding(str, Enum):
    """
    Enum representing the valid encoding formats for public or private keys.

    Attributes:
        BECH32 (str): Encodes the key using Bech32 format.
        HEX (str): Encodes the key using hexadecimal format.
    """

    BECH32 = "bech32"
    HEX = "hex"

    @classmethod
    def from_str(cls, value: str) -> "KeyEncoding":
        """
        Convert a string to a KeyEncoding enum value, case-insensitively.

        Args:
            value (str): The input string, e.g. "bech32" or "HEX".

        Returns:
            KeyEncoding: A corresponding enum value.

        Raises:
            ValueError: If the input is not a valid encoding.
        """
        try:
            return cls(value.lower())
        except ValueError:
            raise ValueError(
                f"Invalid encoding: '{value}'. Must be one of: {[e.value for e in cls]}"
            )


class Namespace(str, Enum):
    """
    Represents a namespace.
    """

    MERCHANT = "com.synvya.merchant"
    GAMER = "com.synvya.gamer"
    OTHER = "com.synvya.other"
    BUSINESS_TYPE = "business.type"

    """Configuration for Pydantic models to use enum values directly."""
    model_config = ConfigDict(use_enum_values=True)


class ProfileType(str, Enum):
    """
    Represents a profile type.
    """

    RETAIL = "retail"
    RESTAURANT = "restaurant"
    SERVICE = "service"
    BUSINESS = "business"
    ENTERTAINMENT = "entertainment"
    OTHER = "other"
    GAMER_DADJOKE = "dad-joke-game"
    OTHER_OTHER = "other"

    """Configuration for Pydantic models to use enum values directly."""
    model_config = ConfigDict(use_enum_values=True)


class ProfileFilter(BaseModel):
    """
    Represents a profile filter.
    """

    namespace: str
    profile_type: ProfileType
    hashtags: List[str]

    def __init__(
        self,
        namespace: str,
        profile_type: ProfileType,
        hashtags: Optional[List[str]] = None,
    ) -> None:
        """
        Initialize a ProfileFilter instance.
        """
        normalized_hashtags = (
            [self._normalize_hashtag(tag) for tag in hashtags] if hashtags else []
        )
        super().__init__(
            namespace=namespace, profile_type=profile_type, hashtags=normalized_hashtags
        )
        self.namespace = namespace
        self.profile_type = profile_type
        self.hashtags = normalized_hashtags

    def to_json(self) -> str:
        """
        Convert the ProfileFilter to a JSON string.
        """
        return json.dumps(self.model_dump())

    @classmethod
    def from_json(cls, json_str: str) -> "ProfileFilter":
        """
        Create a ProfileFilter instance from a JSON string.
        """
        data = json.loads(json_str)
        return cls.model_validate(data)

    @staticmethod
    def _normalize_hashtag(tag: str) -> str:
        """
        Normalize hashtags by converting to lowercase.
        Ensures consistent matching across variations.
        """
        tag = tag.lower()
        # tag = re.sub(r"[\s\-_]+", "", tag)  # Remove spaces, hyphens, underscores
        return tag


class Profile(BaseModel):
    """
    Nostr Profile class.
    Contains public key only.
    """

    PROFILE_URL_PREFIX: ClassVar[str] = "https://primal.net/p/"
    logger: ClassVar[logging.Logger] = logging.getLogger("Profile")

    public_key: str  # stored in hex format
    about: str = ""
    banner: str = ""
    bot: bool = False
    city: str = ""
    country: str = ""
    created_at: int = 0
    display_name: str = ""
    email: str = ""
    hashtags: List[str] = []
    locations: Set[str] = Field(default_factory=set)
    name: str = ""
    namespace: str = ""
    nip05: str = ""
    nip05_validated: bool = False
    picture: str = ""
    phone: str = ""
    profile_type: ProfileType = ProfileType.OTHER_OTHER
    profile_url: str = ""
    state: str = ""
    street: str = ""
    website: str = ""
    zip_code: str = ""
    geohash: str = ""
    environment: Literal["production", "demo"] = "production"

    def __init__(self, public_key: str, **data) -> None:
        """
        Initialize a Profile instance.
        Args:
            public_key: Public key of the Nostr profile in hex or bech32 format.
            **data: Additional data to initialize the Profile instance.

        Returns:
            None
        """
        public_key = PublicKey.parse(public_key).to_hex()
        super().__init__(public_key=public_key, **data)
        self.profile_url = self.PROFILE_URL_PREFIX + public_key

    def add_hashtag(self, hashtag: str) -> None:
        normalized_hashtag = self._normalize_hashtag(hashtag)
        if normalized_hashtag not in self.hashtags:
            self.hashtags.append(normalized_hashtag)

    def add_location(self, location: str) -> None:
        self.locations.add(location)

    def get_about(self) -> str:
        return self.about

    def get_banner(self) -> str:
        return self.banner

    def get_city(self) -> str:
        return self.city

    def get_country(self) -> str:
        return self.country

    def get_created_at(self) -> int:
        return self.created_at

    def get_display_name(self) -> str:
        return self.display_name

    def get_environment(self) -> Literal["production", "demo"]:
        return self.environment

    def get_email(self) -> str:
        return self.email

    def get_hashtags(self) -> List[str]:
        return self.hashtags

    def get_locations(self) -> set[str]:
        return self.locations

    def get_name(self) -> str:
        return self.name

    def get_namespace(self) -> str:
        return self.namespace

    def get_nip05(self) -> str:
        return self.nip05

    def get_phone(self) -> str:
        return self.phone

    def get_picture(self) -> str:
        return self.picture

    def get_profile_type(self) -> ProfileType:
        return self.profile_type

    def get_profile_url(self) -> str:
        return self.profile_url

    def get_public_key(self, encoding: KeyEncoding = KeyEncoding.HEX) -> str:
        """Get the public key of the Nostr profile.

        Args:
            encoding: encoding to use for the public key.
            Must be 'bech32' or 'hex'. Default is 'hex'.

        Returns:
            str: public key of the Nostr profile in the specified encoding

        Raises:
            ValueError: if the encoding is not 'bech32' or 'hex'
        """
        if encoding == KeyEncoding.BECH32:
            return PublicKey.parse(self.public_key).to_bech32()
        if encoding == KeyEncoding.HEX:
            return PublicKey.parse(self.public_key).to_hex()

        raise ValueError("Invalid encoding. Must be 'bech32' or 'hex'.")

    def get_state(self) -> str:
        return self.state

    def get_street(self) -> str:
        return self.street

    def get_website(self) -> str:
        return self.website

    def get_zip_code(self) -> str:
        return self.zip_code

    def get_geohash(self) -> str:
        return self.geohash

    def is_bot(self) -> bool:
        return self.bot

    def is_nip05_validated(self) -> bool:
        return self.nip05_validated

    def matches_filter(self, profile_filter: ProfileFilter) -> bool:
        if self.namespace != profile_filter.namespace:
            return False
        if self.profile_type != profile_filter.type:
            return False
        if not all(hashtag in self.hashtags for hashtag in profile_filter.hashtags):
            return False
        return True

    def set_about(self, about: str) -> None:
        self.about = about

    def set_banner(self, banner: str) -> None:
        self.banner = self._validate_url(banner) if banner else ""

    def set_bot(self, bot: bool) -> None:
        self.bot = bot

    def set_city(self, city: str) -> None:
        self.city = city

    def set_country(self, country: str) -> None:
        self.country = country

    def set_created_at(self, created_at: int) -> None:
        self.created_at = created_at

    def set_display_name(self, display_name: str) -> None:
        self.display_name = display_name

    def set_environment(self, environment: Literal["production", "demo"]) -> None:
        self.environment = environment

    def set_email(self, email: str) -> None:
        self.email = email

    def set_name(self, name: str) -> None:
        self.name = name

    def set_namespace(self, namespace: Namespace | str) -> None:
        if isinstance(namespace, str):
            # Handle empty string by not setting any namespace
            if not namespace:
                self.namespace = ""
                return
            # Convert string to Namespace enum safely
            namespace = Namespace(namespace)
        self.namespace = namespace

    def set_nip05(self, nip05: str) -> None:
        self.nip05 = nip05

    def set_picture(self, picture: str) -> None:
        self.picture = self._validate_url(picture) if picture else ""

    def set_phone(self, phone: str) -> None:
        self.phone = phone

    def set_profile_type(self, profile_type: ProfileType | str) -> None:
        if isinstance(profile_type, str):
            # Convert string to ProfileType enum safely
            profile_type = ProfileType(profile_type)
        self.profile_type = profile_type

    def set_state(self, state: str) -> None:
        self.state = state

    def set_street(self, street: str) -> None:
        self.street = street

    def set_website(self, website: str) -> None:
        self.website = self._validate_url(website) if website else ""

    def set_zip_code(self, zip_code: str) -> None:
        self.zip_code = zip_code

    def set_geohash(self, geohash: str) -> None:
        self.geohash = geohash

    def to_dict(self) -> dict:
        return {
            "about": self.about,
            "banner": self.banner,
            "bot": self.bot,
            "city": self.city,
            "country": self.country,
            "created_at": self.created_at,
            "display_name": self.display_name,
            "environment": self.environment,
            "email": self.email,
            "geohash": self.geohash,
            "hashtags": self.hashtags,
            "locations": list(self.locations),  # Convert set to list
            "name": self.name,
            "namespace": self.namespace,
            "nip05": self.nip05,
            "picture": self.picture,
            "phone": self.phone,
            "profile_url": self.profile_url,
            "public_key": self.public_key,
            "profile_type": self.profile_type,
            "state": self.state,
            "street": self.street,
            "website": self.website,
            "zip_code": self.zip_code,
        }

    def to_json(self) -> str:
        data = {
            "about": self.about,
            "banner": self.banner,
            "bot": self.bot,
            "city": self.city,
            "country": self.country,
            "created_at": self.created_at,
            "display_name": self.display_name,
            "environment": self.environment,
            "email": self.email,
            "geohash": self.geohash,
            "hashtags": self.hashtags,
            "locations": (list(self.locations) if self.locations else []),
            "name": self.name,
            "namespace": self.namespace,
            "nip05": self.nip05,
            "picture": self.picture,
            "phone": self.phone,
            "profile_url": self.profile_url,
            "public_key": self.public_key,
            "profile_type": self.profile_type.value,
            "state": self.state,
            "street": self.street,
            "website": self.website,
            "zip_code": self.zip_code,
        }
        return json.dumps(data)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Profile):
            return False
        return str(self.public_key) == str(other.public_key)

    def __hash__(self) -> int:
        return hash(str(self.public_key))

    async def _fetch_nip05_metadata(self, nip05: str) -> dict:
        """
        Fetch NIP-05 metadata from the well-known URL.

        Args:
            nip05: NIP-05 identifier in the format username@domain

        Returns:
            dict: Parsed JSON response containing metadata

        Raises:
            RuntimeError: if the request fails or returns an error
        """
        if "@" not in nip05:
            raise ValueError("Invalid NIP-05 format. Expected name@domain.")

        name, domain = nip05.split("@")
        url = f"https://{domain}/.well-known/nostr.json?name={name}"

        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(url)
                response.raise_for_status()  # Raise an error for bad responses
                return response.json()
        except httpx.HTTPStatusError as e:
            raise RuntimeError(
                f"Failed to fetch NIP-05 metadata for {nip05}: {e.response.text}"
            ) from e
        except Exception as e:
            raise RuntimeError(
                f"An error occurred while fetching NIP-05 metadata: {e}"
            ) from e

    async def _validate_profile_nip05(self) -> bool:
        if not self.nip05:
            self.logger.error("Profile has no NIP-05")
            return False
        if self.nip05.startswith("@"):
            self.logger.error("Profile NIP-05 must not start with @")
            return False

        try:
            # Extract the local part (username) from the NIP-05 identifier
            local_part = self.nip05.split("@")[0]
            nostr_json = await self._fetch_nip05_metadata(self.nip05)
        except Exception as e:
            self.logger.error("Failed to fetch NIP-05 metadata: %s", e)
            return False

        if "names" in nostr_json:
            for name, public_key in nostr_json["names"].items():
                if (
                    name.lower() == local_part.lower()
                    and public_key == self.get_public_key("hex")
                ):
                    return True
        else:
            return False

    def _validate_url(self, url: str) -> str:
        if not url:
            return ""
        if not url.startswith(("http://", "https://")):
            return ""
        return url

    @classmethod
    @deprecated(
        reason="Method is incomplete and lacks namespace/type/hashtags logic",
        alternative="from_event",
    )
    async def from_metadata(cls, metadata: Metadata, public_key: str) -> "Profile":
        """
        Create a Profile instance from a Metadata object.
        TBD: Add logic to set namespace, type and hashtags from Metadata.
        """
        profile = cls(public_key)
        profile.set_about(metadata.get_about())
        profile.set_banner(metadata.get_banner())
        profile.set_display_name(metadata.get_display_name())
        profile.set_name(metadata.get_name())
        profile.set_nip05(metadata.get_nip05())
        profile.set_picture(metadata.get_picture())
        profile.set_website(metadata.get_website())
        json_bot = metadata.get_custom_field("bot")
        if isinstance(json_bot, JsonValue.BOOL):
            profile.set_bot(json_bot.bool)
        else:
            profile.set_bot(False)

        json_environment = metadata.get_custom_field("environment")
        if isinstance(json_environment, JsonValue.STR):
            profile.set_environment(json_environment.s)
        else:
            profile.set_environment("production")

        try:
            profile.nip05_validated = await profile._validate_profile_nip05()

        except Exception as e:
            profile.logger.error("Failed to validate NIP-05: %s", e)
            profile.nip05_validated = False
        return profile

    @classmethod
    async def from_event(cls, event: Event) -> "Profile":
        """
        Create a Profile instance from a kind:0 Nostr event.

        Args:
            event: kind:0 Nostr event

        Returns:
            Profile: Profile instance

        Raises:
            ValueError: if the event is not a kind:0 Nostr event
        """

        if event.kind() != Kind(0):
            raise ValueError("Event is not a kind:0 Nostr event")

        profile = cls(event.author().to_hex())

        # Process metadata
        metadata = json.loads(event.content())
        profile.set_about(metadata.get("about", ""))
        profile.set_banner(metadata.get("banner", ""))
        profile.set_bot(metadata.get("bot", False))
        # profile.set_city(metadata.get("city", ""))
        # profile.set_country(metadata.get("country", ""))
        profile.set_display_name(metadata.get("display_name", ""))
        profile.set_environment(metadata.get("environment", "production"))
        # profile.set_email(metadata.get("email", ""))
        profile.set_name(metadata.get("name", ""))
        profile.set_nip05(metadata.get("nip05", ""))
        profile.set_picture(metadata.get("picture", ""))
        # profile.set_phone(metadata.get("phone", ""))
        # profile.set_state(metadata.get("state", ""))
        # profile.set_street(metadata.get("street", ""))
        profile.set_website(metadata.get("website", ""))
        # profile.set_zip_code(metadata.get("zip_code", ""))

        profile.set_created_at(event.created_at().as_secs())

        # process tags
        tags = event.tags()

        tag_vector: List[Tag] = tags.to_vec()

        for tag in tag_vector:
            if tag.kind() == TagKind.SINGLE_LETTER(
                SingleLetterTag.lowercase(Alphabet.I)
            ):
                # Extract identity claims from i-tags
                tag_content = tag.content()
                if ":" in tag_content:
                    claim_type, identity = tag_content.split(":", 1)

                    if claim_type == "email":
                        profile.set_email(identity)
                    elif claim_type == "phone":
                        profile.set_phone(identity)
                    elif claim_type == "location":
                        # Parse location string into components
                        # Expected format: street, city, state, country, zip_code
                        # Modify expected format to: street, city, state, zip_code, country
                        location_parts = [part.strip() for part in identity.split(",")]

                        if len(location_parts) >= 1:
                            profile.set_street(location_parts[0])
                        if len(location_parts) >= 2:
                            profile.set_city(location_parts[1])
                        if len(location_parts) >= 3:
                            profile.set_state(location_parts[2])
                        if len(location_parts) >= 4:
                            profile.set_zip_code(location_parts[3])
                        if len(location_parts) >= 5:
                            profile.set_country(location_parts[4])

        hashtag_list = tags.hashtags()
        for hashtag in hashtag_list:
            profile.add_hashtag(hashtag)

        geohash_tag = tags.find(
            TagKind.SINGLE_LETTER(SingleLetterTag.lowercase(Alphabet.G))
        )
        if geohash_tag is not None:
            profile.set_geohash(geohash_tag.content())

        namespace_tag = tags.find(
            TagKind.SINGLE_LETTER(SingleLetterTag.uppercase(Alphabet.L))
        )
        if namespace_tag is not None:
            profile.set_namespace(namespace_tag.content())

        profile_type_tag = tags.find(
            TagKind.SINGLE_LETTER(SingleLetterTag.lowercase(Alphabet.L))
        )
        if profile_type_tag is not None:
            profile.set_profile_type(profile_type_tag.content())

        try:
            profile.nip05_validated = await profile._validate_profile_nip05()
        except Exception as e:
            profile.logger.error("Failed to validate NIP-05: %s", e)
            profile.nip05_validated = False
        profile.logger.debug("Profile: %s", profile)
        return profile

    @classmethod
    def from_json(cls, json_str: str) -> "Profile":
        """
        Create a Profile instance from a JSON string.

        Args:
            json_str (str): JSON string containing profile information.

        Returns:
            Profile: An instance of Profile.
        """
        data = json.loads(json_str)
        profile = cls(public_key=data["public_key"])
        profile.set_about(data.get("about", ""))
        profile.set_banner(data.get("banner", ""))
        profile.set_bot(data.get("bot", False))
        profile.set_city(data.get("city", ""))
        profile.set_country(data.get("country", ""))
        profile.set_created_at(data.get("created_at", 0))
        profile.set_display_name(data.get("display_name", ""))
        profile.set_environment(data.get("environment", "production"))
        profile.set_email(data.get("email", ""))
        profile.set_geohash(data.get("geohash", ""))
        for hashtag in data.get("hashtags", []):
            profile.add_hashtag(hashtag)
        profile.locations = set(data.get("locations", []))
        profile.set_namespace(data.get("namespace", ""))
        profile.set_name(data.get("name", ""))
        profile.set_nip05(data.get("nip05", ""))
        profile.set_picture(data.get("picture", ""))
        profile.set_phone(data.get("phone", ""))
        profile.set_profile_type(data.get("profile_type", ProfileType.OTHER_OTHER))
        profile.set_state(data.get("state", ""))
        profile.set_street(data.get("street", ""))
        profile.set_website(data.get("website", ""))
        profile.set_zip_code(data.get("zip_code", ""))

        return profile

    @staticmethod
    def _normalize_hashtag(tag: str) -> str:
        """
        Normalize hashtags by removing spaces, underscores, and hyphens,
        and converting to lowercase.
        Ensures consistent matching across variations.
        """
        tag = tag.lower()
        tag = re.sub(r"[\s\-_]+", "", tag)  # Remove spaces, hyphens, underscores
        return tag


class NostrKeys(BaseModel):
    """
    NostrKeys is a class that contains a public and private key.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    keys: Keys

    def __init__(self, private_key: Optional[str] = None) -> None:
        """
        Initialize a NostrKeys instance.
        If no private key is provided, a new one will be generated.
        Args:
            private_key: str in bech32 or hex format
        """
        if private_key is None:
            keys = Keys.generate()
        else:
            keys = Keys.parse(private_key)

        super().__init__(keys=keys)

    def get_public_key(self, encoding: KeyEncoding = KeyEncoding.BECH32) -> str:
        """
        Get the public key.

        Args:
            encoding (KeyEncoding, optional): The encoding format for the public key.

        Returns:
            str: public key

        Raises:
            ValueError: If the encoding is not 'bech32' or 'hex'.
        """
        if isinstance(encoding, str):
            try:
                encoding = KeyEncoding(encoding.lower())
            except ValueError as e:
                raise ValueError(
                    f"Invalid encoding. Must be one of: {[e.value for e in KeyEncoding]}"
                ) from e

        match encoding:
            case KeyEncoding.BECH32:
                return self.keys.public_key().to_bech32()
            case KeyEncoding.HEX:
                return self.keys.public_key().to_hex()

    def get_private_key(self, encoding: KeyEncoding = KeyEncoding.BECH32) -> str:
        """
        Get the private key.

        Args:
            encoding (KeyEncoding, optional): The encoding format for the public key.

        Returns:
            str: private key

        Raises:
            ValueError: If the encoding is not 'bech32' or 'hex'.
        """
        if isinstance(encoding, str):
            try:
                encoding = KeyEncoding(encoding.lower())
            except ValueError as e:
                raise ValueError(
                    f"Invalid encoding. Must be one of: {[e.value for e in KeyEncoding]}"
                ) from e

        match encoding:
            case KeyEncoding.BECH32:
                return self.keys.secret_key().to_bech32()
            case KeyEncoding.HEX:
                return self.keys.secret_key().to_hex()

    def to_json(self, encoding: KeyEncoding = KeyEncoding.BECH32) -> str:
        """
        Returns a JSON representation of the NostrKeys object.

        The resulting JSON contains both the public and private keys in bech32 format.

        Returns:
            str: JSON string with keys.
        """
        if isinstance(encoding, str):
            try:
                encoding = KeyEncoding(encoding.lower())
            except ValueError as e:
                raise ValueError(
                    f"Invalid encoding. Must be one of: {[e.value for e in KeyEncoding]}"
                ) from e

        match encoding:
            case KeyEncoding.BECH32:
                return json.dumps(
                    {
                        "public_key": self.keys.public_key().to_bech32(),
                        "private_key": self.keys.secret_key().to_bech32(),
                    }
                )
            case KeyEncoding.HEX:
                return json.dumps(
                    {
                        "public_key": self.keys.public_key().to_hex(),
                        "private_key": self.keys.secret_key().to_hex(),
                    }
                )

    def __str__(self) -> str:
        """Return a string representation of the NostrKeys object using bech32 encoding."""
        return f"Public_key: {self.keys.public_key().to_bech32()} \nPrivate_key: {self.keys.secret_key().to_bech32()}"

    @classmethod
    @deprecated(
        reason="Use new NostrKeys constructor instead.",
        alternative="NostrKeys(private_key=private_key)",
    )
    def from_private_key(cls, private_key: str) -> "NostrKeys":
        """Create a NostrKeys object from a private key."""
        return cls(private_key)

    @classmethod
    def derive_public_key(
        cls, private_key: str, encoding: KeyEncoding = KeyEncoding.BECH32
    ) -> str:
        """
        Class method to parse a private key and return a public key
        in bech32 or hex format.

        Args:
            private_key (str): The private key to derive the public key from.
            encoding (KeyEncoding): The encoding to use for the public key.

        Returns:
            str: The public key in the specified encoding.

        Raises:
            ValueError: If the encoding is not 'bech32' or 'hex'.
        """
        if isinstance(encoding, str):
            try:
                encoding = KeyEncoding(encoding.lower())
            except ValueError as e:
                raise ValueError(
                    f"Invalid encoding. Must be one of: {[e.value for e in KeyEncoding]}"
                ) from e
        match encoding:
            case KeyEncoding.BECH32:
                return Keys.parse(private_key).public_key().to_bech32()
            case KeyEncoding.HEX:
                return Keys.parse(private_key).public_key().to_hex()


def _extract_nostr_tags(event: "Event") -> List[List[str]]:
    """
    Extract the raw tag arrays from a nostr_sdk Event.
    """
    tags_data: List[List[str]] = []

    as_json = getattr(event, "as_json", None)
    if callable(as_json):
        try:
            raw_event = as_json()
            event_dict = (
                json.loads(raw_event) if isinstance(raw_event, str) else raw_event
            )
            raw_tags = event_dict.get("tags", [])
            for tag in raw_tags:
                if isinstance(tag, list):
                    tags_data.append([str(item) for item in tag])
            if tags_data:
                return tags_data
        except Exception:
            tags_data = []

    try:
        tags_obj = event.tags()
        to_vec = getattr(tags_obj, "to_vec", None)
        if callable(to_vec):
            for tag in to_vec():
                tag_values = None
                as_vec = getattr(tag, "as_vec", None)
                if callable(as_vec):
                    tag_values = as_vec()
                elif hasattr(tag, "as_json"):
                    try:
                        maybe_json = tag.as_json()
                        tag_values = (
                            json.loads(maybe_json)
                            if isinstance(maybe_json, str)
                            else maybe_json
                        )
                    except Exception:
                        tag_values = None
                if not tag_values or not isinstance(tag_values, list):
                    continue
                tags_data.append([str(item) for item in tag_values])
    except Exception:
        return tags_data

    return tags_data


class ClassifiedListing(BaseModel):
    """
    Represents a marketplace classified listing (kind 30402).
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: str
    title: str
    description: str = ""
    price_amount: Optional[float] = None
    price_currency: Optional[str] = None
    price_frequency: Optional[str] = None
    listing_type: str = "simple"
    listing_format: str = "digital"
    visibility: str = "on-sale"
    stock: Optional[int] = None
    summary: Optional[str] = None
    images: List[Dict[str, Optional[str]]] = Field(default_factory=list)
    specs: List[List[str]] = Field(default_factory=list)
    weight_value: Optional[str] = None
    weight_unit: Optional[str] = None
    dimensions_value: Optional[str] = None
    dimensions_unit: Optional[str] = None
    location: Optional[str] = None
    geohash: Optional[str] = None
    categories: List[str] = Field(default_factory=list)
    shipping_options: List[Dict[str, Optional[str]]] = Field(default_factory=list)
    collections: List[str] = Field(default_factory=list)
    seller: str = ""  # hex format

    def set_seller(self, seller: str) -> None:
        self.seller = PublicKey.parse(seller).to_hex()

    def get_seller(self) -> str:
        return self.seller

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "title": self.title,
            "description": self.description,
            "price": {
                "amount": self.price_amount,
                "currency": self.price_currency,
                "frequency": self.price_frequency,
            },
            "listing_type": self.listing_type,
            "listing_format": self.listing_format,
            "visibility": self.visibility,
            "stock": self.stock,
            "summary": self.summary,
            "images": self.images,
            "specs": self.specs,
            "weight": {"value": self.weight_value, "unit": self.weight_unit},
            "dimensions": {
                "value": self.dimensions_value,
                "unit": self.dimensions_unit,
            },
            "location": self.location,
            "geohash": self.geohash,
            "categories": self.categories,
            "shipping_options": self.shipping_options,
            "collections": self.collections,
            "seller": self.seller,
        }

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_event(cls, event: "Event") -> "ClassifiedListing":
        tags_data = _extract_nostr_tags(event)

        def find_first(name: str) -> Optional[List[str]]:
            for tag in tags_data:
                if tag and tag[0] == name:
                    return tag
            return None

        def find_all(name: str) -> List[List[str]]:
            return [tag for tag in tags_data if tag and tag[0] == name]

        id_tag = find_first("d")
        title_tag = find_first("title")
        price_tag = find_first("price")
        type_tag = find_first("type")
        visibility_tag = find_first("visibility")
        stock_tag = find_first("stock")
        summary_tag = find_first("summary")
        weight_tag = find_first("weight")
        dim_tag = find_first("dim")
        location_tag = find_first("location")
        geohash_tag = find_first("g")

        identifier = id_tag[1] if id_tag and len(id_tag) > 1 else ""
        title = title_tag[1] if title_tag and len(title_tag) > 1 else ""

        price_amount: Optional[float] = None
        price_currency: Optional[str] = None
        price_frequency: Optional[str] = None
        if price_tag and len(price_tag) > 2:
            try:
                price_amount = float(price_tag[1])
            except (TypeError, ValueError):
                price_amount = None
            price_currency = price_tag[2]
            if len(price_tag) > 3:
                price_frequency = price_tag[3]

        listing_type = type_tag[1] if type_tag and len(type_tag) > 1 else "simple"
        listing_format = type_tag[2] if type_tag and len(type_tag) > 2 else "digital"
        visibility = (
            visibility_tag[1]
            if visibility_tag and len(visibility_tag) > 1
            else "on-sale"
        )

        stock: Optional[int] = None
        if stock_tag and len(stock_tag) > 1:
            try:
                stock = int(stock_tag[1])
            except (TypeError, ValueError):
                stock = None

        summary = summary_tag[1] if summary_tag and len(summary_tag) > 1 else None
        weight_value = weight_tag[1] if weight_tag and len(weight_tag) > 1 else None
        weight_unit = weight_tag[2] if weight_tag and len(weight_tag) > 2 else None
        dimensions_value = dim_tag[1] if dim_tag and len(dim_tag) > 1 else None
        dimensions_unit = dim_tag[2] if dim_tag and len(dim_tag) > 2 else None
        location = location_tag[1] if location_tag and len(location_tag) > 1 else None
        geohash = geohash_tag[1] if geohash_tag and len(geohash_tag) > 1 else None

        images: List[Dict[str, Optional[str]]] = []
        for image_tag in find_all("image"):
            image_entry: Dict[str, Optional[str]] = {}
            if len(image_tag) > 1:
                image_entry["url"] = image_tag[1]
            if len(image_tag) > 2:
                image_entry["dimensions"] = image_tag[2]
            if len(image_tag) > 3:
                image_entry["order"] = image_tag[3]
            images.append(image_entry)

        specs = [
            [spec_tag[1], spec_tag[2]]
            for spec_tag in find_all("spec")
            if len(spec_tag) > 2
        ]

        categories = [cat_tag[1] for cat_tag in find_all("t") if len(cat_tag) > 1]

        shipping_options = []
        for opt_tag in find_all("shipping_option"):
            entry: Dict[str, Optional[str]] = {}
            if len(opt_tag) > 1:
                entry["reference"] = opt_tag[1]
            if len(opt_tag) > 2:
                entry["extra_cost"] = opt_tag[2]
            shipping_options.append(entry)

        collections = [
            col_tag[1]
            for col_tag in find_all("a")
            if len(col_tag) > 1 and col_tag[1].startswith("30405:")
        ]

        try:
            seller = event.author().to_hex()
        except Exception:
            seller = ""

        return cls(
            id=identifier,
            title=title,
            description=event.content() or "",
            price_amount=price_amount,
            price_currency=price_currency,
            price_frequency=price_frequency,
            listing_type=listing_type,
            listing_format=listing_format,
            visibility=visibility,
            stock=stock,
            summary=summary,
            images=images,
            specs=specs,
            weight_value=weight_value,
            weight_unit=weight_unit,
            dimensions_value=dimensions_value,
            dimensions_unit=dimensions_unit,
            location=location,
            geohash=geohash,
            categories=categories,
            shipping_options=shipping_options,
            collections=collections,
            seller=seller,
        )


class Collection(BaseModel):
    """
    Represents a product collection (kind 30405).
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: str
    title: str
    description: str = ""
    product_references: List[str] = Field(default_factory=list)
    image: Optional[str] = None
    summary: Optional[str] = None
    location: Optional[str] = None
    geohash: Optional[str] = None
    shipping_options: List[Dict[str, Optional[str]]] = Field(default_factory=list)
    author: str = ""

    def set_author(self, author: str) -> None:
        self.author = author

    def get_author(self) -> str:
        return self.author

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "title": self.title,
            "description": self.description,
            "product_references": self.product_references,
            "image": self.image,
            "summary": self.summary,
            "location": self.location,
            "geohash": self.geohash,
            "shipping_options": self.shipping_options,
            "author": self.author,
        }

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_event(cls, event: "Event") -> "Collection":
        tags_data = _extract_nostr_tags(event)

        def find_first(name: str) -> Optional[List[str]]:
            for tag in tags_data:
                if tag and tag[0] == name:
                    return tag
            return None

        def find_all(name: str) -> List[List[str]]:
            return [tag for tag in tags_data if tag and tag[0] == name]

        id_tag = find_first("d")
        title_tag = find_first("title")
        image_tag = find_first("image")
        summary_tag = find_first("summary")
        location_tag = find_first("location")
        geohash_tag = find_first("g")

        identifier = id_tag[1] if id_tag and len(id_tag) > 1 else ""
        title = title_tag[1] if title_tag and len(title_tag) > 1 else ""

        product_references = [
            prod_tag[1]
            for prod_tag in find_all("a")
            if len(prod_tag) > 1 and prod_tag[1].startswith("30402:")
        ]

        shipping_options = []
        for opt_tag in find_all("shipping_option"):
            entry: Dict[str, Optional[str]] = {}
            if len(opt_tag) > 1:
                entry["reference"] = opt_tag[1]
            if len(opt_tag) > 2:
                entry["extra_cost"] = opt_tag[2]
            shipping_options.append(entry)

        try:
            author = event.author().to_bech32()
        except Exception:
            author = event.author().to_hex()

        return cls(
            id=identifier,
            title=title,
            description=event.content() or "",
            product_references=product_references,
            image=image_tag[1] if image_tag and len(image_tag) > 1 else None,
            summary=summary_tag[1] if summary_tag and len(summary_tag) > 1 else None,
            location=(
                location_tag[1] if location_tag and len(location_tag) > 1 else None
            ),
            geohash=geohash_tag[1] if geohash_tag and len(geohash_tag) > 1 else None,
            shipping_options=shipping_options,
            author=author,
        )


class ProductShippingCost(BaseModel):
    psc_id: str
    psc_cost: float

    def __init__(self, psc_id: str, psc_cost: float) -> None:
        super().__init__(psc_id=psc_id, psc_cost=psc_cost)
        self.psc_id = psc_id
        self.psc_cost = psc_cost

    def get_id(self) -> str:
        return self.psc_id

    def get_cost(self) -> float:
        return self.psc_cost

    def set_id(self, psc_id: str) -> None:
        self.psc_id = psc_id

    def set_cost(self, psc_cost: float) -> None:
        self.psc_cost = psc_cost

    def to_dict(self) -> dict:
        return {"id": self.psc_id, "cost": self.psc_cost}

    def to_json(self) -> str:
        """Returns a JSON representation of the ProductShippingCost object."""
        return json.dumps(self.to_dict())

    def __str__(self) -> str:
        return f"ID: {self.psc_id} Cost: {self.psc_cost}"


class StallShippingMethod(BaseModel):
    """
    Represents a shipping method for a stall.
    """

    ssm_id: str
    ssm_cost: float
    ssm_name: str
    ssm_regions: List[str] = Field(default_factory=list)

    def __init__(
        self,
        ssm_id: str,
        ssm_cost: float,
        ssm_name: str,
        ssm_regions: Optional[List[str]] = None,
    ) -> None:
        super().__init__(
            ssm_id=ssm_id,
            ssm_cost=ssm_cost,
            ssm_name=ssm_name,
            ssm_regions=ssm_regions if ssm_regions is not None else [],
        )
        self.ssm_id = ssm_id
        self.ssm_cost = ssm_cost
        self.ssm_name = ssm_name
        self.ssm_regions = ssm_regions if ssm_regions is not None else []

    def get_id(self) -> str:
        return self.ssm_id

    def get_cost(self) -> float:
        return self.ssm_cost

    def get_name(self) -> str:
        return self.ssm_name

    def get_regions(self) -> List[str]:
        return self.ssm_regions

    def set_id(self, ssm_id: str) -> None:
        self.ssm_id = ssm_id

    def set_cost(self, ssm_cost: float) -> None:
        self.ssm_cost = ssm_cost

    def set_name(self, ssm_name: str) -> None:
        self.ssm_name = ssm_name

    def set_regions(self, ssm_regions: List[str]) -> None:
        self.ssm_regions = ssm_regions

    def to_dict(self) -> dict:
        return {
            "id": self.ssm_id,
            "cost": self.ssm_cost,
            "name": self.ssm_name,
            "regions": self.ssm_regions,
        }

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    def __str__(self) -> str:
        return (
            f"ID: {self.ssm_id} "
            f"Cost: {self.ssm_cost} "
            f"Name: {self.ssm_name} "
            f"Regions: {self.ssm_regions}"
        )


class Product(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: str
    stall_id: str
    name: str
    description: str
    images: List[str]
    currency: str
    price: float
    quantity: int
    shipping: List[ProductShippingCost]
    categories: List[str] = Field(default_factory=list)
    specs: List[List[str]] = Field(default_factory=list)
    seller: str  # hex

    def set_seller(self, seller: str) -> None:
        """
        Set the seller of the product.
        Use it to set the seller after creating a Product using the
        @classmethod from_product_data() since ProductData does not contain
        the seller's public key.

        Args:
            seller: str in hex format
        """
        self.seller = seller

    def get_seller(self) -> str:
        """Get the seller of the product.

        Returns:
            str: seller of the product in hex format
        """
        return self.seller

    @classmethod
    def from_product_data(cls, product_data: "ProductData") -> "Product":
        shipping_costs = []
        for ship in product_data.shipping:
            if isinstance(ship, dict):
                shipping_costs.append(
                    ProductShippingCost(psc_id=ship["id"], psc_cost=ship["cost"])
                )
            else:
                shipping_costs.append(
                    ProductShippingCost(psc_id=ship.id, psc_cost=ship.cost)
                )

        # Handle specs - ensure it's a list
        specs = []
        if product_data.specs is not None:
            if isinstance(product_data.specs, dict):
                # Convert dict to list of lists if needed
                specs = [[k, v] for k, v in product_data.specs.items()]
            elif isinstance(product_data.specs, list):
                specs = product_data.specs

        return cls(
            id=product_data.id,
            stall_id=product_data.stall_id,
            name=product_data.name,
            description=product_data.description,
            images=product_data.images,
            currency=product_data.currency,
            price=product_data.price,
            quantity=product_data.quantity,
            shipping=shipping_costs,
            categories=(
                product_data.categories if product_data.categories is not None else []
            ),
            specs=specs,
            seller="",
        )

    def to_product_data(self) -> "ProductData":
        try:
            # Convert self.shipping from List[ProductShippingCost] to List[ShippingCost]
            shipping_costs = [
                ShippingCost(id=shipping.psc_id, cost=shipping.psc_cost)
                for shipping in self.shipping
            ]

            return ProductData(
                id=self.id,
                stall_id=self.stall_id,
                name=self.name,
                description=self.description,
                images=self.images,
                currency=self.currency,
                price=self.price,
                quantity=self.quantity,
                shipping=shipping_costs,  # Use the converted shipping costs
                categories=self.categories,
                specs=self.specs,
            )
        except Exception as e:
            logging.error("Failed to convert to ProductData: %s", e)
            logging.error("Shipping data: %s", self.shipping)
            raise

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the Product.

        Returns:
            dict: dictionary representation of the Product
        """
        # Use the to_dict method of ProductShippingCost for serialization
        shipping_dicts = [
            {"id": shipping.psc_id, "cost": shipping.psc_cost}
            for shipping in self.shipping
        ]

        return {
            "id": self.id,
            "stall_id": self.stall_id,
            "name": self.name,
            "description": self.description,
            "images": self.images,
            "currency": self.currency,
            "price": self.price,
            "quantity": self.quantity,
            "shipping": shipping_dicts,  # Use the serialized shipping costs
            "categories": self.categories,
            "specs": self.specs,
            "seller": self.seller,
        }

    def to_json(self) -> str:
        """
        Returns a JSON string representation of the Product.

        Returns:
            str: JSON string representation of the Product
        """
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> "Product":
        """
        Create a Product instance from a JSON string.

        Args:
            json_str (str): JSON string containing product information.

        Returns:
            Product: An instance of Product.
        """
        data = json.loads(json_str)
        shipping_costs = [
            ProductShippingCost(psc_id=ship["id"], psc_cost=ship["cost"])
            for ship in data.get("shipping", [])
        ]
        return cls(
            id=data["id"],
            stall_id=data["stall_id"],
            name=data["name"],
            description=data["description"],
            images=data.get("images", []),
            currency=data["currency"],
            price=data["price"],
            quantity=data["quantity"],
            shipping=shipping_costs,
            categories=data.get("categories", []),
            specs=data.get("specs", []),
            seller=data["seller"],
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Product):
            return False
        return str(self.id) == str(other.id)


class Stall(BaseModel):
    """
    Stall represents a NIP-15 stall.
    TBD: NIP-15 does not have a geohash field. Add logic to retrieve geohash from
    somewhere else when using the from_stall_data() class constructor.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    id: str
    name: str
    description: str
    currency: str
    shipping: List[StallShippingMethod]
    geohash: Optional[str] = None

    def get_geohash(self) -> str:
        return self.geohash

    def set_geohash(self, geohash: str) -> None:
        self.geohash = geohash

    def to_dict(self) -> dict:
        # Use the to_dict method of StallShippingMethod for serialization
        shipping_dicts = [shipping.to_dict() for shipping in self.shipping]

        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "currency": self.currency,
            "shipping": shipping_dicts,  # Use the serialized shipping methods
            "geohash": self.geohash,
        }

    def to_json(self) -> str:
        """
        Returns a JSON string representation of the Stall.

        Returns:
            str: JSON string representation of the Stall
        """
        return json.dumps(self.to_dict())

    def to_stall_data(self) -> "StallData":
        # Convert self.shipping from List[StallShippingMethod] to List[ShippingMethod]
        shipping_methods = [
            ShippingMethod(id=shipping.ssm_id, cost=shipping.ssm_cost)
            .name(shipping.ssm_name)
            .regions(shipping.ssm_regions)
            for shipping in self.shipping
        ]

        return StallData(
            self.id,
            self.name,
            self.description,
            self.currency,
            # self.shipping,  # No conversion needed
            shipping_methods,
        )

    @classmethod
    def from_json(cls, stall_content: str) -> "Stall":
        # Handle empty strings
        if not stall_content or stall_content.isspace():
            return cls(
                id="unknown",
                name="Unknown Stall",
                description="",
                currency="USD",
                shipping=[],
            )

        # Parse the JSON string into a data structure
        try:
            data = json.loads(stall_content)
        except json.JSONDecodeError:
            # Return default stall if JSON parsing fails
            return cls(
                id="unknown",
                name="Unknown Stall",
                description="",
                currency="USD",
                shipping=[],
            )

        # If data is an empty list, create a default Stall
        if isinstance(data, list):
            if not data:  # Empty list
                return cls(
                    id="unknown",
                    name="Unknown Stall",
                    description="",
                    currency="USD",
                    shipping=[],
                )
            # If it's a non-empty list, use the first item
            data = data[0] if isinstance(data[0], dict) else {}

        # Ensure data is a dictionary at this point
        if not isinstance(data, dict):
            # If data is neither a list nor a dict, create a default stall
            return cls(
                id="unknown",
                name="Unknown Stall",
                description="",
                currency="USD",
                shipping=[],
            )

        # Create a list of StallShippingMethod from the shipping data
        shipping_methods = []
        # Handle stalls that might not have shipping data
        if "shipping" in data:
            for shipping in data.get("shipping", []):
                if not isinstance(shipping, dict):
                    continue
                # Handle case where regions might be None
                regions = shipping.get("regions", [])
                try:
                    shipping_methods.append(
                        StallShippingMethod(
                            ssm_id=shipping.get("id", "unknown"),
                            ssm_cost=float(shipping.get("cost", 0.0)),
                            ssm_name=shipping.get("name", ""),
                            ssm_regions=regions if regions is not None else [],
                        )
                    )
                except (ValueError, TypeError):
                    # Skip this shipping method if any errors occur
                    continue

        # Handle missing required fields with defaults
        return cls(
            id=data.get("id", "unknown"),
            name=data.get("name", data.get("title", "")),  # Try title as fallback
            description=data.get("description", ""),
            currency=data.get("currency", "USD"),  # Default to USD
            shipping=shipping_methods,
            geohash=data.get("geohash", None),
        )

        # ------------------------------------------------------------------------------ #
        # Delegation (NIP-26) Support
        # ------------------------------------------------------------------------------ #

        # class Delegation(BaseModel):
        #     """
        #     NIP‑26 delegation wrapper.

        #     A merchant signs a *kind 30078* event delegating publishing rights to the
        #     server.  This helper parses that event, validates its signature and
        #     provides convenience checks for downstream publishing code.
        #     """

        #     author: str  # Merchant pubkey (bech32 or hex)
        #     conditions: str  # Raw query string e.g. "kind=30078&expires_at=…"
        #     sig: str  # Merchant signature
        #     tag: list[str]  # Complete ["delegation", …] tag to re‑attach
        #     created_at: int  # Delegation creation (unix ts)
        #     expires_at: int  # Expiry (unix ts)
        #     allowed_kinds: Set[int]  # Kinds we may publish

        #     # ------------------------------------------------------------------ #
        #     # Construction helpers
        #     # ------------------------------------------------------------------ #
        #     @classmethod
        #     def parse(cls, raw: str | dict) -> "Delegation":
        #         """
        #         Convert raw JSON (str or dict) of a *kind 30078* event into a validated
        #         Delegation instance.

        #         Raises:
        #             ValueError - on wrong kind or bad signature.
        #         """
        #         evt = raw if isinstance(raw, dict) else json.loads(raw)

        #         # Basic integrity checks
        #         if evt.get("kind") != 30078:
        #             raise ValueError("Event is not a delegation (kind 30078)")

        #         # Verify sig using nostr‑sdk
        #         event_obj = Event.from_json(json.dumps(evt))
        #         if not event_obj.verify():
        #             raise ValueError("Invalid delegation signature")

        #         # Pull out the delegation tag parts
        #         tags = evt.get("tags", [])

        #         # Find the delegation tag (should be ["delegation", delegatee, conditions, token])
        #         delegation_tag = None
        #         for tag in tags:
        #             if len(tag) >= 4 and tag[0] == "delegation":
        #                 delegation_tag = tag
        #                 break

        #         if delegation_tag is None:
        #             raise ValueError("Delegation tag missing")

        #         cond_str = delegation_tag[2]  # "kind=0,1,30023&created_at<1751565393"
        #         cond_map = {}

        #         # Parse conditions more flexibly to handle different formats
        #         for condition in cond_str.split("&"):
        #             if "=" in condition:
        #                 k, v = condition.split("=", 1)
        #                 cond_map[k] = v
        #             elif "<" in condition:
        #                 # Handle created_at<timestamp format
        #                 k, v = condition.split("<", 1)
        #                 if k == "created_at":
        #                     cond_map["expires_at"] = (
        #                         v  # created_at<timestamp means expires at timestamp
        #                     )

        #         allowed = set()
        #         if "kind" in cond_map:
        #             allowed = {int(k) for k in cond_map["kind"].split(",") if k.isdigit()}

        #         created = int(cond_map.get("created_at", evt["created_at"]))
        #         expires = int(cond_map.get("expires_at", created))

        #         return cls(
        #             author=evt["pubkey"],
        #             conditions=cond_str,
        #             sig=evt["sig"],
        #             tag=delegation_tag,
        #             created_at=created,
        #             expires_at=expires,
        #             allowed_kinds=allowed,
        #         )

        # ------------------------------------------------------------------ #
        # Validation helpers
        # ------------------------------------------------------------------ #
        # def validate_event(self, event: Event) -> None:
        #     """
        #     Ensure *event* is publishable under this delegation.

        #     Raises:
        #         ValueError - if kind not allowed or delegation expired.
        #     """
        #     event_kind_value = event.kind().as_u16()  # Convert Kind to integer
        #     if event_kind_value not in self.allowed_kinds and self.allowed_kinds:
        #         raise ValueError("Event kind not allowed by delegation")

        #     now_ts = int(datetime.now(timezone.utc).timestamp())
        #     if now_ts > self.expires_at:
        #         raise ValueError("Delegation expired")

        # # Convenience: ready‑made tag to append before publish()
        # @property
        # def delegation_tag(self) -> list[str]:
        #     return self.tag

        # @property
        # def delegatee(self) -> str:
        #     """
        #     Get the delegatee's public key from the delegation tag.

        #     Note: According to NIP-26, the delegatee public key is NOT stored in the
        #     delegation tag. The delegatee is the entity that will use this delegation
        #     to publish events. This property returns empty string as the delegatee
        #     must be determined from context.

        #     Returns:
        #         str: Empty string (delegatee not stored in delegation tag)
        #     """
        #     return ""  # Delegatee not stored in delegation tag per NIP-26

        # @property
        # def delegator(self) -> str:
        #     """
        #     Get the delegator's public key (same as author).

        #     Returns:
        #         str: Public key of the delegator (entity granting delegation rights)
        #     """
        # return self.author



"""
Synvya SDK: Tools for a Nostr agentic ecosystem
"""

import importlib.metadata
import logging

# Import main classes to make them available at package level
from .models import (
    ClassifiedListing,
    Collection,
    KeyEncoding,
    Namespace,
    NostrKeys,
    Product,
    ProductShippingCost,
    Profile,
    ProfileFilter,
    ProfileType,
    Stall,
    StallShippingMethod,
)
from .nostr import NostrClient, generate_keys, verify_signature

# Import version from pyproject.toml at runtime
try:
    __version__ = importlib.metadata.version("synvya_sdk")
except importlib.metadata.PackageNotFoundError:
    logging.warning("Package 'synvya_sdk' not found. Falling back to 'unknown'.")
    __version__ = "unknown"
except ImportError:
    logging.warning("importlib.metadata is not available. Falling back to 'unknown'.")
    __version__ = "unknown"

# Define What is Exposed at the Package Level
__all__ = [
    "NostrClient",
    "generate_keys",
    "Profile",
    "ProfileFilter",
    "ProfileType",
    "Namespace",
    "ProductShippingCost",
    "StallShippingMethod",
    "Product",
    "Stall",
    "NostrKeys",
    "verify_signature",
    "generate_keys",
    "ClassifiedListing",
    "Collection",
]



"""
Module implementing the DadJokeTools Toolkit for Agno agents.

Publisher sends a joke request to a joker:
{
    "role": "publisher",
    "content": "Please send me a pg dad joke..."
}

Joker receives the joke request and sends a joke to the publisher:
{
    "role": "joker",
    "content": "Why did the chicken cross the road?"
}
"""

import json
import secrets
from typing import List, Optional, Union

from pydantic import ConfigDict

from synvya_sdk import Namespace, NostrClient, Profile, ProfileFilter, ProfileType

try:
    from agno.tools import Toolkit
    from agno.utils.log import logger
except ImportError as exc:
    raise ImportError(
        "Package `agno` not installed. Please install using `pip install agno`"
    ) from exc


class DadJokeGamerTools(Toolkit):
    """
    DadJokeTools is a toolkit that allows an agent to play the Dad Joke game.
    """

    model_config = ConfigDict(
        arbitrary_types_allowed=True, extra="allow", validate_assignment=True
    )

    _instances_from_create: set[int] = set()

    def __init__(
        self,
        name: str,
        relays: Union[str, List[str]],
        private_key: str,
        _from_create: bool = False,
    ) -> None:
        """
        Initialize the DadJokeGamerTools object.

        Args:
            name: Name of the gamer
            relays: Nostr relay(s) that the client will connect to. Can be a single URL string or a list of URLs.
            private_key: Private key for the client in hex or bech32 format
            _from_create: Internal flag to ensure proper initialization flow
        """
        if not _from_create:
            raise RuntimeError(
                "DadJokeGamerTools must be created using the create() method"
            )

        # Track instance ID
        self._instance_id = id(self)
        DadJokeGamerTools._instances_from_create.add(self._instance_id)

        super().__init__(name=name)

        # Set include_tools to None to allow all tools to be registered
        self.include_tools = None

        # Convert single relay to list for consistent handling
        self.relays: List[str] = [relays] if isinstance(relays, str) else relays
        self.private_key: str = private_key
        self.nostr_client: Optional[NostrClient] = None
        self.profile: Optional[Profile] = None
        self.joker_public_key: Optional[str] = None

        # Register methods
        # Publisher
        self.register(self.async_find_joker)
        self.register(self.async_listen_for_joke)
        self.register(self.async_publish_joke)
        self.register(self.async_request_joke)

        # Joker
        self.register(self.async_listen_for_joke_request)
        self.register(self.async_submit_joke)

    def __del__(self) -> None:
        """
        Delete the DadJokeGamerTools instance.
        """
        if hasattr(self, "_instance_id"):
            DadJokeGamerTools._instances_from_create.discard(self._instance_id)

    @classmethod
    async def create(
        cls, name: str, relays: Union[str, List[str]], private_key: str
    ) -> "DadJokeGamerTools":
        """
        Asynchronous factory method for proper initialization.
        Use instead of the __init__ method.

        Args:
            name: Name of the gamer
            relays: Nostr relay(s) that the client will connect to. Can be a single URL string or a list of URLs.
            private_key: Private key for the client in hex or bech32 format

        Returns:
            DadJokeGamerTools: An initialized DadJokeGamerTools instance
        """
        instance = cls(name, relays, private_key, _from_create=True)

        # Initialize NostrClient asynchronously
        instance.nostr_client = await NostrClient.create(relays, private_key)
        instance.profile = await instance.nostr_client.async_get_profile()
        instance.nostr_client.set_logging_level(logger.getEffectiveLevel())
        return instance

    async def async_find_joker(self) -> str:
        """
        Finds all jokers in the network and selects one at random.
        Jokers are defined as Profiles meeting the following criteria:
        - Must have a validated NIP-05 identity.
        - Must have the metadata field `bot` set to true.
        - The kind:0 event must include the label `dad-joke-game` with
        namespace `com.synvya.gamer`
        - The kind:0 event must include the hashtag `joker`

        Returns:
            str: JSON string containing the bech32 encoded public key of the joker
        """
        if self.nostr_client is None:
            raise RuntimeError("NostrClient not initialized. Call create() first.")

        NostrClient.logger.info("Finding jokers")
        joker_filter = ProfileFilter(
            namespace=Namespace.GAMER,
            profile_type=ProfileType.GAMER_DADJOKE,
            hashtags=["joker"],
        )

        agents = await self.nostr_client.async_get_agents(joker_filter)

        response = {
            "status": "error",
            "message": "No valid jokers found",
        }

        if agents:
            tries = 0
            while tries < 10:
                # Use secrets for secure random selection
                agents_list = list(agents)
                selected_joker: Profile = agents_list[
                    secrets.randbelow(len(agents_list))
                ]
                if selected_joker.is_nip05_validated() and selected_joker.is_bot():
                    response = {
                        "status": "success",
                        "joker": selected_joker.get_public_key(encoding="bech32"),
                    }
                    self.joker_public_key = selected_joker.get_public_key(
                        encoding="bech32"
                    )
                    break
                tries += 1
        return json.dumps(response)

    async def async_listen_for_joke(self, timeout: int = 60) -> str:
        """
        Listen for a joke.

        Expecting a kind:14 message containing a JSON
        object with the following fields:
        - role: "joker"
        - content: "The joke"
        """
        if self.nostr_client is None:
            raise RuntimeError("NostrClient not initialized. Call create() first.")

        NostrClient.logger.info("Listening for a joke")
        try:
            message = await self.nostr_client.async_receive_message(timeout)
            message_dict = json.loads(message)
            # let's make sure the joke came from the joker we request the joke from
            if message_dict.get("sender") != self.joker_public_key:
                return json.dumps({"status": "error", "message": "Unknown message"})

            message_type = message_dict.get("type")
            if message_type == "kind:14":
                content_dict = json.loads(message_dict.get("content"))
                if content_dict.get("role") == "joker":
                    return json.dumps(
                        {
                            "status": "success",
                            "joke": content_dict.get("content"),
                            "joker": message_dict.get("sender"),
                        }
                    )
        except Exception as e:
            return json.dumps(
                {
                    "status": "error",
                    "message": str(e),
                }
            )
        return json.dumps({"status": "error", "message": "No joke received."})

    async def async_publish_joke(self, joke: str, joker_public_key: str) -> str:
        """
        Publish a joke as a kind:1 event

        Args:
            joke: joke to publish

        Returns:
            str: JSON string containing the status of the publication
        """
        if self.nostr_client is None:
            raise RuntimeError("NostrClient not initialized. Call create() first.")

        NostrClient.logger.info("Publishing a joke")
        try:
            text = f"Dad Joke from @{joker_public_key}:\n {joke}"
            await self.nostr_client.async_publish_note(text)
            return json.dumps(
                {
                    "status": "success",
                    "message": "Joke published",
                }
            )
        except Exception as e:
            return json.dumps(
                {
                    "status": "error",
                    "message": str(e),
                }
            )

    async def async_request_joke(self, joker_public_key: str) -> str:
        """
        Request a joke from a joker.

        Args:
            joker_public_key: bech32 encoded public key of the joker

        Returns:
            str: JSON string containing the status of the request
        """
        if self.nostr_client is None:
            raise RuntimeError("NostrClient not initialized. Call create() first.")

        NostrClient.logger.info("Requesting a joke")
        message = json.dumps(
            {
                "role": "publisher",
                "content": "Please send me a pg dad joke...",
            }
        )

        await self.nostr_client.async_send_message(
            "kind:14",
            joker_public_key,
            message,
        )

        return json.dumps(
            {
                "status": "success",
                "message": "Joke requested",
            }
        )

    async def async_listen_for_joke_request(self, timeout: int = 7200) -> str:
        """
        Listen for a joke request.
        """
        if self.nostr_client is None:
            raise RuntimeError("NostrClient not initialized. Call create() first.")

        NostrClient.logger.info("Listening for a joke request")
        try:
            message = await self.nostr_client.async_receive_message(timeout)
            message_dict = json.loads(message)

            # let's make sure the request came from a publisher
            if message_dict.get("type") == "kind:14":
                sender = message_dict.get("sender")
                profile = await self.nostr_client.async_get_profile(sender)
                if (
                    profile.get_namespace() == Namespace.GAMER
                    and profile.get_profile_type() == ProfileType.GAMER_DADJOKE
                    and "publisher" in profile.get_hashtags()
                ):
                    message_content = json.loads(message_dict.get("content"))
                    if message_content.get("role") == "publisher":
                        return json.dumps(
                            {
                                "status": "success",
                                "message": "Joke request received",
                                "publisher": sender,
                            }
                        )
        except Exception as e:
            return json.dumps({"status": "error", "message": str(e)})
        return json.dumps({"status": "error", "message": "No joke request received."})

    async def async_submit_joke(self, joke: str, publisher: str) -> str:
        """
        Submit a joke.
        """
        if self.nostr_client is None:
            raise RuntimeError("NostrClient not initialized. Call create() first.")

        NostrClient.logger.info("Submitting a joke")
        try:
            await self.nostr_client.async_send_message(
                "kind:14",
                publisher,
                json.dumps({"role": "joker", "content": joke}),
            )
            return json.dumps({"status": "success", "message": "Joke submitted"})
        except Exception as e:
            return json.dumps({"status": "error", "message": str(e)})

    async def async_set_profile(self, profile: Profile) -> str:
        """
        Sets the profile used by the Toolkit.
        The profile is also published to the Nostr network.

        Returns:
            str: id of the event publishing the profile

        Raises:
            RuntimeError: if it can't publish the event
        """
        if self.nostr_client is None:
            raise RuntimeError("NostrClient not initialized. Call create() first.")

        try:
            result: str = await self.nostr_client.async_set_profile(profile)
            return result
        except RuntimeError as e:
            logger.error("Unable to publish the profile: %s", e)
            raise RuntimeError(f"Unable to publish the profile: {e}") from e

    def get_profile(self) -> str:
        """
        Get the merchant profile in JSON format

        Returns:
            str: merchant profile in JSON format
        """
        if self.profile is None:
            raise RuntimeError("Profile not initialized. Call create() first.")
        return json.dumps(self.profile.to_json())
